###############################################################################################################################
# The goal here is to implement Eq. 9.4 from the LSST community WP, which defines our FoM, and create plots.
#
# Humna Awan: humna.awan@rutgers.edu
#
###############################################################################################################################
__all__ = (
    "get_fsky",
    "get_theory_spectra",
    "get_outdir_name",
    "return_cls",
    "calc_os_bias_err",
    "get_fom",
    "os_bias_overplots",
    "os_bias_overplots_diff_dbs",
)

import datetime
import os
from collections import OrderedDict

import healpy as hp
import matplotlib.pyplot as plt
import numpy as np

from rubin_sim.maf.maf_contrib.lss_obs_strategy.constants_for_pipeline import power_law_const_a


###############################################################################################################################
# calculate fsky for a bundle
def get_fsky(outdir, band="i", print_fsky=True):
    """

    Calculate the fraction of the sky observed in a survey. The data must have been saved as
    .npz files in the given output directory. The method only looks at the mask of the data array.

    Filenames should be in the format: <whatever>_<band>_<dither_strategy>.npz

    Parameters
    -------------------
    outdir : str
        name of the output directory where the data-to-look-at is.
    band: str
        band to consider. Default: 'i'
    print_fsky: `bool`
        set to True if want to print( out fsky. Default: True

    """
    filenames = [f for f in os.listdir(outdir) if any([f.endswith("npz")])]
    fsky = {}
    for filename in filenames:
        # print('Reading in %s for fsky'%filename)
        dither_strategy = filename.split("%s_" % band)[1].split(".npz")[0]
        data = np.load("%s/%s" % (outdir, filename))
        # total number of pixels in the sky
        tot_pix = float(len(data["mask"]))
        in_survey_pix = float(len(np.where(data["mask"] == False)[0]))
        fsky[dither_strategy] = in_survey_pix / tot_pix
        if print_fsky:
            print("%s fsky: %s\n" % (dither_strategy, fsky[dither_strategy]))
    return fsky


###############################################################################################################################
def get_theory_spectra(mock_data_path, mag_cut=25.6, plot_spectra=True, nside=256):
    """

    Return the data for the five redshift bins, read from the files containing the
    with BAO galaxy power spectra from Hu Zhan.

    Parameters
    -------------------
    mock_data_path : `str`
        path to the folder with the theory spectra
    mag_cut: `float`
        r-band magnitude cut as the identifer in the filename from Hu.
        allowed options: 24.0, 25.6, 27.5. Default: 25.6
    plot_spectra : `bool`
        set to True if want to plot out the skymaps. Default: True
    nside : `int`
        HEALpix resolution parameter. Default: 256

    Returns
    -------
    ell : `np.ndarray`
        array containing the ells
    w_bao_cls : `dict`
        keys = zbin_tags; data = spectra (pixelized for specified nside)
    surf_num_density : `float`
        surface number density in 1/Sr
    """
    # read in the galaxy power spectra with the BAO
    filename = "%s/cls015-200z_r%s.bins" % (mock_data_path, mag_cut)
    print("\nReading in %s for theory cls." % filename)
    shot_noise_data = np.genfromtxt(
        filename
    )  # last column = surface number density for each bin in 1/(sq arcmin)

    filename = "%s/cls015-200z_r%s" % (mock_data_path, mag_cut)
    print("Reading in %s for theory cls." % filename)
    all_data = np.genfromtxt(filename)

    # set up to read the data
    ell = []
    w_bao_cls = OrderedDict()
    w_bao_cls["0.15<z<0.37"] = []
    w_bao_cls["0.37<z<0.66"] = []
    w_bao_cls["0.66<z<1.0"] = []
    w_bao_cls["1.0<z<1.5"] = []
    w_bao_cls["1.5<z<2.0"] = []
    surf_num_density = OrderedDict()

    # read in the cls
    for i in range(len(all_data)):
        ell.append(all_data[i][0])
        w_bao_cls["0.15<z<0.37"].append(all_data[i][1])
        w_bao_cls["0.37<z<0.66"].append(all_data[i][3])
        w_bao_cls["0.66<z<1.0"].append(all_data[i][5])
        w_bao_cls["1.0<z<1.5"].append(all_data[i][7])
        w_bao_cls["1.5<z<2.0"].append(all_data[i][9])

    # read in the surface number density and convert
    for j, key in enumerate(w_bao_cls.keys()):
        surf_num_density[key] = np.array(shot_noise_data[j][5] * 1.18 * 10**7)  # 1/arcmin^2 to 1/Sr

    # want Cl*W^2 where W is the pixel window function
    wl_nside = hp.sphtfunc.pixwin(nside=nside)

    # account for Hu's ells not starting with 0 but with ell=2
    wl_nside = wl_nside[2:]
    lmax = len(wl_nside)
    ell = np.array(ell[0:lmax])
    # pixelize the spectra
    for key in w_bao_cls:
        w_bao_cls[key] = np.array(w_bao_cls[key][0:lmax]) * (wl_nside**2)

    if plot_spectra:
        plt.clf()
        for key in w_bao_cls:
            plt.plot(
                ell,
                w_bao_cls[key] * ell * (ell + 1) / (2 * np.pi),
                linewidth=1.5,
                label=key,
            )
        plt.legend(bbox_to_anchor=(1, 1))
        plt.xlabel("$\ell$")
        plt.ylabel("$\ell(\ell+1)C_\ell/2\pi$")
        plt.ticklabel_format(axis="y", style="sci", scilimits=(-2, 2))
        plt.show()

    return ell, w_bao_cls, surf_num_density


###############################################################################################################################
# return the outputDir name where the relevant c_ells are.
def get_outdir_name(
    band,
    nside,
    pixel_radius,
    yr_cutoff,
    zbin,
    mag_cut_i,
    run_name,
    poisson_noise,
    zero_pt,
):
    """

    Return the output directory name where the cls for deltaN/N would be, given the input parameters.
    We assume that dust extinction is always added and counts are always normalized.

    Returns: [outdir, yr_tag, zbin_tag, poisson_tag, zero_pt_tag]

    Parameters
    -------------------
      * band: str: band to get the output directory name for.
                   Options: 'u', 'g', 'r', 'i'
      * nside: int: HEALpix resolution parameter.
      * pixel_radius: int: number of pixels to mask along the shallow border
      * yr_cutoff: int: year cut to restrict analysis to only a subset of the survey.
                         Must range from 1 to 9, or None for the full survey analysis (10 yrs).
      * zbin: str: options include '0.<z<0.15', '0.15<z<0.37', '0.37<z<0.66, '0.66<z<1.0',
                          '1.0<z<1.5', '1.5<z<2.0', '2.0<z<2.5', '2.5<z<3.0','3.0<z<3.5', '3.5<z<4.0',
                          'all' for no redshift restriction (i.e. 0.<z<4.0)
      * mag_cut_i: float: upper limit on i-band magnitude when calculating the galaxy counts. will deal
                         with color correction automatically (for u,g,r bands ..) depending on the band
      * run_name: str: run name tag to identify the output of specified OpSim output.
                      Since new OpSim outputs have different columns, the run_name for enigma_1189 **must**
                      be 'enigma1189'; can be anything for other outputs, e.g. 'minion1016'
      * poisson_noise: bool: set to True to consider the case where poisson noise is added to  galaxy
                                   counts after border masking (and the incorporation of calibration errors).
      * zero_pt: bool: set to True to consider the case where photometric calibration errors were incorporated.

    """
    # check to make sure redshift bin is ok.
    allowed_zbins = power_law_const_a.keys()
    if zbin not in allowed_zbins:
        raise ValueError("ERROR: Invalid redshift bin. Input bin can only be among %s\n" % (allowed_zbins))

    # set up the tags.
    dust_tag = "withDustExtinction"

    zero_pt_tag = ""
    if zero_pt:
        zero_pt_tag = "with0ptErrors"
    else:
        zero_pt_tag = "no0ptErrors"

    if yr_cutoff is not None:
        yr_tag = "%syearCut" % yr_cutoff
    else:
        yr_tag = "fullSurveyPeriod"

    zbin_tag = zbin
    if zbin == "all":
        zbin_tag = "allRedshiftData"

    if poisson_noise:
        poisson_tag = "withPoissonNoise"
    else:
        poisson_tag = "noPoissonNoise"

    norm_tag = "normalizedGalaxyCounts"

    # account for color corrections.
    mag_cut = {}
    mag_cut["i"] = mag_cut_i
    mag_cut["r"] = float("%.1f" % (mag_cut["i"] + 0.4))
    mag_cut["g"] = float("%.1f" % (mag_cut["r"] + 0.4))
    mag_cut["u"] = float("%.1f" % (mag_cut["g"] + 0.4))

    outdir = "artificialStructure_%s_nside%s_%spixelRadiusForMasking_%s_%s" % (
        poisson_tag,
        nside,
        pixel_radius,
        zero_pt_tag,
        dust_tag,
    )
    outdir += "_%s<%s_%s_%s_%s_%s_directory/" % (
        band,
        mag_cut[band],
        run_name,
        yr_tag,
        zbin_tag,
        norm_tag,
    )

    return [outdir, yr_tag, zbin_tag, poisson_tag, zero_pt_tag]


###############################################################################################################################
# create a function that returns cls for a given band, nside, pixel_radius
def return_cls(path, outdir, band, specified_dith=None):
    """

    Get the cls from .npy files in path+outdir folder for a specified filter band.

    Returns the data in the form of a dictonary with dither strategies as keys.

    Parameters
    -------------------
      * path: str: path to the main directory where directories for the outputs from
                   artificialStructure are saved.
      * outdir: str: name of the directory where the cls are situated.
      * band: str: band to consider. Options: 'u', 'g', 'r', 'i', 'z', 'y'
      * specified_dith: list of str: list of the names (strings) of the dither strategies to consider, e.g.
                                    if want to plot only NoDither, specified_dith_only= ['NoDither']
    """
    if specified_dith is None:
        consider_all_npy = True
    else:
        consider_all_npy = False
    # get the filename
    filenames = [f for f in os.listdir("%s%s" % (path, outdir)) if any([f.endswith("npy")])]
    # read in the cls
    c_ells = {}
    for filename in filenames:
        if consider_all_npy:
            dither_strategy = filename.split("%s_" % band)[1].split(".npy")[0]
            c_ells[dither_strategy] = np.load("%s%s/%s" % (path, outdir, filename))
        else:
            for dith in specified_dith:
                if filename.__contains__(dith):
                    dither_strategy = filename.split("%s_" % band)[1].split(".npy")[0]
                    c_ells[dither_strategy] = np.load("%s%s/%s" % (path, outdir, filename))
    if specified_dith is not None:
        for dith in specified_dith:
            if dith not in c_ells:
                raise ValueError("Cls for %s are not found in %s." % (dith, outdir))
    return c_ells


###############################################################################################################################
def calc_os_bias_err(c_ells):
    """

    Calculate the OS bias (as an average across the specified bands) and the uncertainty in the
    bias (as the std across the cls from thes specified bands).

    Returns two dictionaries: [bias, bias_err]

    Parameters
    ----------
    c_ells: dictionary
        bands as keys, mapping the cls corresponding to the bands.

    """
    bias, bias_err = {}, {}
    band_keys = list(c_ells.keys())
    for dith in c_ells[band_keys[0]]:  # loop over each dith strategy
        temp_avg, temp_err = [], []
        for data_index in range(len(c_ells[band_keys[0]][dith])):  # loop over each C_ell-value
            row = []
            for band in band_keys:
                row.append(c_ells[band][dith][data_index])  # compiles the C_ell value for each band
            temp_avg.append(np.mean(row))
            temp_err.append(np.std(row))
        bias[dith] = temp_avg
        bias_err[dith] = temp_err
    return [bias, bias_err]


###############################################################################################################################
def get_stat_floor(
    ell_arr,
    zbin,
    w_bao_cls_zbin,
    surf_num_density_zbin,
    dither_strategy,
    fsky,
    with_shot_noise=True,
):
    # returns the sqrt of cosmic variance = statistical floor
    prefactor = np.sqrt(2.0 / ((2 * ell_arr + 1) * fsky))
    if with_shot_noise:
        return prefactor * (w_bao_cls_zbin + (1.0 / surf_num_density_zbin))
    else:
        return prefactor * (w_bao_cls_zbin)


###############################################################################################################################
def get_fom(
    ell_min,
    ell_max,
    ell_for_bias_err,
    bias_err,
    ell_stat_floor,
    floor_with_shot_noise,
    floor_wo_shot_noise,
):
    """

    Calculate the FoM based on the bias uncertaity and statistical floor. Returns the FoM (float).

    Parameters
    -------------------
      * ell_min: int: minimum ell-value to consider over which the FoM is calculated.
      * ell_max: int: maximum ell-value to consider over which the FoM is calculated.
      * ell_for_bias_err: array: ell-array corresponding to the bias_err array.
      * bias_err: array: array containing the bias_err for each ell.
      * ell_stat_floor: array: ell-array corresponding to the get_stat_floor array.
      * floor_with_shot_noise: array: array containing the statistical floor (for each ell)  with shot noise contribution.
      * floor_wo_shot_noise: array: array containing the statistical floor (for each ell) without shot noise contribution.

    """
    # need to adjust for the ell's not always starting with 0.
    osbias = np.array(bias_err)[ell_min - ell_for_bias_err[0] : ell_max - ell_for_bias_err[0] + 1]
    floor_with_shot_noise = floor_with_shot_noise[
        int(ell_min - ell_stat_floor[0]) : int(ell_max - ell_stat_floor[0] + 1)
    ]
    floor_wo_shot_noise = floor_wo_shot_noise[
        int(ell_min - ell_stat_floor[0]) : int(ell_max - ell_stat_floor[0] + 1)
    ]

    l_good = np.arange(ell_min, ell_max + 1)
    num_sq = np.sum(floor_wo_shot_noise**2)
    denom_sq = np.sum(floor_with_shot_noise**2 + osbias**2)

    return np.sqrt(num_sq / denom_sq)


###############################################################################################################################
def os_bias_overplots(
    out_dir,
    data_paths,
    lim_mags_i,
    legend_labels,
    fsky_dith_dict,
    fsky_best,
    mock_data_path,
    run_name,
    theory_lim_mag,
    specified_dith_only=None,
    run_name_filetag=None,
    ell_min=100,
    ell_max=300,
    lmax=500,
    filters=["u", "g", "r", "i"],
    nside=256,
    pixel_radius=14,
    yr_cutoff=None,
    zbin="0.66<z<1.0",
    poisson_noise=False,
    zero_pt=True,
    plot_interms=False,
    color_dict=None,
    ylim_min=None,
    ylim_max=None,
    show_plot=False,
    suptitle=None,
    file_append=None,
):
    """

    Calculate/plot the OS bias uncertainty and the statistical floor for the specified redshift bin.

    Could vary the dither strategies, but the data should be from the same OpSim run. The title of
    of each panel in final plot will be <dither strategy>, and each panel can have OS bias
    uncertainity from many different galaxy catalogs. Panel legends will specify the redshift bin
    and the magnitude cut.

    Parameters
    -------------------
      * out_dir: str: output directory where the output plots will be saved; a folder named
                      'os_bias_overplots' will be created in the directory, if its not there already.
      * data_paths: list of strings: list of strings containing the paths where the artificialStructure data will be
                                     found for the filters specified.
      * lim_mags_i: list of floats: list of the i-band magnitude cuts to get the data for.
      * legend_labels: list of strings: list of the 'tags' for each case; will be used in the legends. e.g. if
                                          lim_mags_i=[24.0, 25.3], legend_labels could be ['i<24.0', 'i<25.3']
                                          or ['r<24.4', 'i<25.7'] if want to label stuff with r-band limits.
      * fsky_dith_dict: dict: dictionary containing the fraction of sky covered; keys should be the dither strategies.
                                The function get_fsky outputs the right dictionary.
      * fsky_best: float: best fsky for the survey to compare everything relative to.
      * mock_data_path: str: path to the mock data to consider
      * run_name: str: run name tag to identify the output of specified OpSim output.
      * theory_lim_mag: float: magnitude cut as the identifer in the filename from Hu.
                               Allowed options: 24.0, 25.6, 27.5
      * specified_dith_only: list of string: list of the names (strings) of the dither strategies to consider, e.g.
                                               if want to plot only NoDither, specified_dith_only= ['NoDither']. If
                                               nothing is specified, all the dither strategies will be considered
                                               (based on the npy files available for the runs). Default: None
      * run_name_filtag: str: run name file tag. Default: None
      * filters: list of strings: list containing the bands (in strings) to be used to calculate the OS bias
                                  and its error. should contain AT LEAST two bands.
                                  e.g. if filters=['g', 'r'], OS bias (at every ell) will be calculated as the
                                  mean across g and r cls, while the bias error (at every ell) will be calculated
                                  as the std. dev. across g and r cls.
                                  Default: ['u', 'g', 'r', 'i']
      * nside: int: HEALpix resolution parameter. Default: 256
      * pixel_radius: int: number of pixels to mask along the shallow border. Default: 14
      * yr_cutoff: int: year cut to restrict analysis to only a subset of the survey.
                        Must range from 1 to 9, or None for the full survey analysis (10 yrs).
                        Default: None
      * zbin: str: options include '0.15<z<0.37', '0.37<z<0.66, '0.66<z<1.0', '1.0<z<1.5', '1.5<z<2.0'
                   Default: '0.66<z<1.0'
      * poisson_noise: bool: set to True to consider the case where poisson noise is added to galaxy counts
                                   after border masking (and the incorporation of calibration errors).
                                   Default: False
      * zero_pt: bool: set to True to consider the case where 0pt calibration errors were incorporated.
                          Default: True
      * plot_interms: bool: set to True to plot intermediate plots, e.g. BAO data. Default: False
      * color_dict: dict: color dictionary; keys should be the indentifiers provided. Default: None
                    **** Please note that in-built colors are for only a few indentifiers:
                        'r<24.0'], 'r<25.7','r<27.5', 'r<22.0', 'i<24.0', 'i<25.3','i<27.5', 'i<22.' ******
      * ylim_min: float: lower y-lim for the final plots. Defaut: None
      * ylim_max: float: upper y-lim for the final plots. Defaut: None
      * show_plot: bool: set to True if want to display the plot (aside from saving it). Default: False
      * suptitle: str: title to plot. Default: None
      * file_append: str: optional string to append to the saved plot

    """
    # check to make sure redshift bin is ok.
    allowed_zbins = list(power_law_const_a.keys()) + ["all"]
    if zbin not in allowed_zbins:
        raise ValueError("Invalid redshift bin. Input bin can only be among %s\n" % (allowed_zbins))

    # check to make sure we have at least two bands to calculate the bias uncertainty.
    if len(filters) < 2:
        raise ValueError(
            "Need at least two filters to calculate bias uncertainty. Currently given only: %s\n" % filters
        )

    # all is ok. proceed.
    tot_cases = len(data_paths)

    # get the outdir address for each 'case' and each band
    outdir_all = {}
    for i in range(tot_cases):
        outdir = {}
        for band in filters:
            out, yr_tag, zbin_tag, poisson_tag, zero_pt_tag = get_outdir_name(
                band,
                nside,
                pixel_radius,
                yr_cutoff,
                zbin,
                lim_mags_i[i],
                run_name,
                poisson_noise,
                zero_pt,
            )
            outdir[band] = "%s/cls_DeltaByN/" % out
        outdir_all[legend_labels[i]] = outdir

    # get the cls and calculate the OS bias and error.
    osbias_all, osbias_err_all = {}, {}
    for i in range(tot_cases):
        outdir = outdir_all[legend_labels[i]]

        c_ells = {}
        for band in filters:
            c_ells[band] = return_cls(
                data_paths[i], outdir[band], band, specified_dith=specified_dith_only
            )  # get the c_ells
        osbias, osbias_err = calc_os_bias_err(c_ells)
        osbias_all[legend_labels[i]] = osbias
        osbias_err_all[legend_labels[i]] = osbias_err

    # get the data to calculate the statistical floor.
    ell, w_bao_cls, surf_num_density = get_theory_spectra(
        mock_data_path=mock_data_path,
        mag_cut=theory_lim_mag,
        plot_spectra=plot_interms,
        nside=nside,
    )
    ########################################################################################################
    # set the directory
    outdir = "os_bias_overplots"
    if not os.path.exists("%s/%s" % (out_dir, outdir)):
        os.makedirs("%s/%s" % (out_dir, outdir))

    in_built_colors = {}
    in_built_colors["r<24.0"] = "r"
    in_built_colors["r<25.7"] = "b"
    in_built_colors["r<27.5"] = "g"
    in_built_colors["r<22.0"] = "m"
    in_built_colors["i<24.0"] = "r"
    in_built_colors["i<25.3"] = "b"
    in_built_colors["i<27.5"] = "g"
    in_built_colors["i<22.0"] = "m"

    if color_dict is None:
        colors = in_built_colors
    else:
        colors = color_dict

    # figure out max how many panels to create
    diths_to_consider = []
    if specified_dith_only is not None:
        diths_to_consider = specified_dith_only
        max_entries = len(diths_to_consider)
    else:
        diths_to_consider = colors.keys()
        max_entries = 0
        for identifier in legend_labels:
            max_entries = max(max_entries, len(list(osbias_err_all[identifier].keys())))

    ncols = 2
    if max_entries == 1:
        ncols = 1
    nrows = int(np.ceil(max_entries / ncols))

    # set up the figure
    plt.clf()
    fig, ax = plt.subplots(nrows, ncols)
    fig.subplots_adjust(hspace=0.4)
    row, col = 0, 0
    # run over the keys
    for dith in diths_to_consider:
        # ----------------------------------------------------------------------------------------
        for i in range(tot_cases):
            # ----------------------------------------------------------------------------------------
            osbias_err = osbias_err_all[legend_labels[i]]
            if dith in osbias_err.keys():
                # ----------------------------------------------------------------------------------------
                if nrows == 1:
                    if ncols == 1:
                        axis = ax
                    else:
                        axis = ax[col]
                else:
                    axis = ax[row, col]

                # ----------------------------------------------------------------------------------------
                # calcuate the floors with and without shot noise
                if zbin not in w_bao_cls:
                    raise ValueError("Invalid redshift bin: %s" % zbin)
                else:
                    # get the floor with shot noise
                    floor_with_eta = get_stat_floor(
                        ell_arr=ell,
                        zbin=zbin,
                        w_bao_cls_zbin=w_bao_cls[zbin],
                        surf_num_density_zbin=surf_num_density[zbin],
                        dither_strategy=dith,
                        fsky=fsky_dith_dict[dith],
                        with_shot_noise=True,
                    )
                    # get the "best" floor for the fom calculation
                    floor_no_eta = get_stat_floor(
                        ell_arr=ell,
                        zbin=zbin,
                        w_bao_cls_zbin=w_bao_cls[zbin],
                        surf_num_density_zbin=surf_num_density[zbin],
                        dither_strategy=dith,
                        fsky=fsky_best,
                        with_shot_noise=False,
                    )
                # calculate the fo_m
                l = np.arange(np.size(osbias_err[dith]))
                fo_m = get_fom(
                    ell_min,
                    ell_max,
                    l,
                    osbias_err[dith],
                    ell,
                    floor_with_eta,
                    floor_no_eta,
                )
                # ----------------------------------------------------------------------------------------
                if i == 0:
                    # plot the floor with shot noise
                    axis.plot(
                        ell,
                        floor_with_eta,
                        color="k",
                        lw=2.0,
                        label="$\Delta$C$_\ell$: $%s$" % zbin,
                    )

                # set up the legend
                add_leg = ""
                if i == 0:
                    add_leg = "$\mathrm{\sigma_{C_{\ell,OS}}}$: "
                else:
                    add_leg = "         "

                # plot the bias error
                splits = legend_labels[i].split("<")
                axis.plot(
                    l,
                    osbias_err[dith],
                    color=colors[legend_labels[i]],
                    label=r"%s$%s<%s$ ; fo_m: %.6f" % (add_leg, splits[0], splits[1], fo_m),
                )
                # ----------------------------------------------------------------------------------------
                # set up the details of the plot
                if (ylim_min is not None) & (ylim_max is not None):
                    axis.set_ylim(ylim_min, ylim_max)
                else:
                    ax.set_ylim(0, 0.00001)
                axis.set_title(dith)
                axis.set_xlabel(r"$\ell$")
                axis.set_xlim(0, lmax)
                if tot_cases > 4:
                    leg = axis.legend(
                        labelspacing=0.001,
                    )
                else:
                    leg = axis.legend()
                axis.ticklabel_format(axis="y", style="sci", scilimits=(-2, 2))
                for legobj in leg.legendHandles:
                    legobj.set_linewidth(2.0)
        row += 1
        if row > nrows - 1:
            row = 0
            col += 1
    # ----------------------------------------------------------------------------------------
    if suptitle is not None:
        plt.suptitle(suptitle)
    # turn stuff off if have odd number of panels
    if (max_entries % 2 != 0) and (max_entries > 1):  # i.e. have odd number of diths
        ax[nrows - 1, ncols - 1].axis("off")

    width = 20
    fig.set_size_inches(width, int(nrows * 30 / 7.0))

    # set up the filename
    dith_tag = ""
    if specified_dith_only is not None:
        dith_tag = "_%sdiths" % max_entries
    if run_name_filetag is None:
        run_name_filetag = run_name
    date_tag = datetime.date.isoformat(datetime.date.today())
    bias_type_tag = "".join(str(x) for x in filters)
    ell_tag = "_%s<ell<%s" % (ell_min, ell_max)
    if file_append is None:
        file_append = ""

    filename = "%s_OSbiaserr_%s_%s%s_" % (
        date_tag,
        bias_type_tag,
        run_name_filetag,
        dith_tag,
    )
    if tot_cases == 1:
        filename += "%s_" % (legend_labels[0])
    else:
        filename += "%s-magcuts_" % (tot_cases)
    filename += "th-r<%s_%s_%s_%s_%s%s%s.png" % (
        theory_lim_mag,
        zbin_tag,
        yr_tag,
        poisson_tag,
        zero_pt_tag,
        ell_tag,
        file_append,
    )
    # save the plot
    plt.savefig("%s/%s/%s" % (out_dir, outdir, filename), format="png", bbox_inches="tight")

    print("\nSaved %s" % filename)
    if show_plot:
        plt.show()
    else:
        plt.close("all")


###############################################################################################################################
def os_bias_overplots_diff_dbs(
    out_dir,
    data_path,
    run_names,
    legend_labels,
    fsky_dict,
    fsky_best,
    mock_data_path,
    theory_lim_mag,
    lim_mag_i,
    ell_min=100,
    ell_max=300,
    lmax=500,
    specified_dith_only=None,
    filters=["u", "g", "r", "i"],
    nside=256,
    pixel_radius=14,
    yr_cutoff=None,
    zbin="0.66<z<1.0",
    poisson_noise=False,
    zero_pt=True,
    plot_interms=False,
    color_dict=None,
    ylim_min=None,
    ylim_max=None,
    show_plot=False,
    suptitle=None,
    file_append=None,
):
    """

    Calculate/plot the OS bias uncertainty and the statistical floor for the specified redshift bin.

    Could vary the dither strategies, but the data should be for the same magnitude cut. The title of
    of each panel in final plot will be "<dither strategy>, and each panel can have
    OS bias uncertainity from many different cadences. Panel legends will specify the redshift bin
    and OpSim output tag.

    Parameters
    -------------------
      * out_dir: str: main directory where the output plots should be saved; a folder named
                      'os_bias_overplots' will be created in the directory, if its not there already.
      * data_path: str: path to the artificialStructure data.
      * run_names: list of str: list for run name tags to identify the output of specified OpSim outputs.
      * legend_labels: list of strings: list of the 'tags' for each case; will be used in the legends. e.g. if
                                          run_names=['enigma1189', 'minion1016'], legend_labels could be
                                          ['enigma_1189', 'minion_1016'].
      * fsky_dict: dict: dictionary of the dictionaries containing the fraction of sky covered for each of the
                         cadences. The keys should match the identifier; fsky_dict[indentifiers[:]] should have
                         the dither strategies as the keys.
      * fsky_best: float: best fsky for the survey to compare everything relative to.
      * mock_data_path: str: path to the mock data to consider
      * theory_lim_mag: float: magnitude cut as the identifer in the filename from Hu.
                               Allowed options: 24.0, 25.6, 27.5
      * lim_mag_i: float: i-band magnitude cut to get the data for.
      * specified_dith_only: list of string: list of the names (strings) of the dither strategies to consider, e.g.
                                           if want to plot only NoDither, specified_dith_only=['NoDither']. If
                                           nothing is specified, all the dither strategies will be considered
                                           (based on the npy files available for the runs). Default: None
      * filters: list of strings: list containing the bands (in strings) to be used to calculate the OS bias
                                  and its error. should contain AT LEAST two bands.
                                  e.g. if filters=['g', 'r'], OS bias (at every ell) will be calculated as the
                                  mean across g and r c_ells, while the bias error (at every ell) will be calculated
                                  as the std. dev. across g and r c_ells.
                                  Default: ['u', 'g', 'r', 'i']
      * nside: int: HEALpix resolution parameter. Default: 256
      * pixel_radius: int: number of pixels to mask along the shallow border. Default: 14
      * yr_cutoff: int: year cut to restrict analysis to only a subset of the survey.
                         Must range from 1 to 9, or None for the full survey analysis (10 yrs).
                         Default: None
      * zbin: str: options include '0.15<z<0.37', '0.37<z<0.66, '0.66<z<1.0', '1.0<z<1.5', '1.5<z<2.0'
                           Default: '0.66<z<1.0'
      * poisson_noise: bool: set to True to consider the case where poisson noise is added to galaxy counts
                                   after border masking (and the incorporation of calibration errors).
                                   Default: False
      * zero_pt: bool: set to True to consider the case where 0pt calibration errors were incorporated.
                          Default: True
      * plot_interms: bool: set to True to plot intermediate plots, e.g. BAO data. Default: False
      * color_dict: dict: color dictionary; keys should be the indentifiers provided. Default: None
                    **** Please note that in-built colors are for only a few indentifiers:
                            minion1016, minion1020, kraken1043 ******
      * ylim_min: float: lower y-lim for the final plots. Defaut: None
      * ylim_max: float: upper y-lim for the final plots. Defaut: None
      * show_plot: bool: set to True if want to display the plot (aside from saving it). Default: False
      * suptitle: str: title to the plot. Default: None
      * file_append: str: optional string to append to the saved plot

    """
    # check to make sure redshift bin is ok.
    allowed_zbins = list(power_law_const_a.keys()) + ["all"]
    if zbin not in allowed_zbins:
        raise ValueError("Invalid redshift bin. Input bin can only be among %s\n" % (allowed_zbins))

    # check to make sure we have at least two bands to calculate the bias uncertainty.
    if len(filters) < 2:
        raise ValueError(
            "Need at least two filters to calculate bias uncertainty. Currently given only: %s\n" % filters
        )

    # all is ok. proceed.
    tot_cases = len(run_names)

    # get the outdir address for each 'case' and each band
    outdir_all = {}
    for i in range(tot_cases):
        outdir = {}
        for band in filters:
            out, yr_tag, zbin_tag, poisson_tag, zero_pt_tag = get_outdir_name(
                band,
                nside,
                pixel_radius,
                yr_cutoff,
                zbin,
                lim_mag_i,
                run_names[i],
                poisson_noise,
                zero_pt,
            )
            outdir[band] = "%s/cls_DeltaByN/" % out
        outdir_all[run_names[i]] = outdir

    # get the cls and calculate the OS bias and error.
    osbias_all, osbias_err_all = {}, {}
    for i in range(tot_cases):
        outdir = outdir_all[run_names[i]]
        c_ells = {}
        for band in filters:
            c_ells[band] = return_cls(
                data_path, outdir[band], band, specified_dith=specified_dith_only
            )  # get the c_ells
        osbias, osbias_err = calc_os_bias_err(c_ells)
        osbias_all[run_names[i]] = osbias
        osbias_err_all[run_names[i]] = osbias_err

    # print stuff
    print(
        "MagCuts: i<%s\nRedshift bin: %s, %s, %s, %s"
        % (lim_mag_i, zbin_tag, yr_tag, poisson_tag, zero_pt_tag)
    )

    # get the data to calculate the statistical floor.
    ell, w_bao_cls, surf_num_density = get_theory_spectra(
        mock_data_path=mock_data_path,
        mag_cut=theory_lim_mag,
        plot_spectra=plot_interms,
        nside=nside,
    )

    ########################################################################################################
    # set the directory
    outdir = "os_bias_overplots"
    if not os.path.exists("%s/%s" % (out_dir, outdir)):
        os.makedirs("%s/%s" % (out_dir, outdir))

    in_built_colors = {}
    in_built_colors["minion1016"] = "r"
    in_built_colors["minion1020"] = "b"
    in_built_colors["kraken1043"] = "g"

    if color_dict is None:
        colors = in_built_colors
    else:
        colors = color_dict

    # figure out how many panels we need
    max_entries = 0
    if specified_dith_only is not None:
        max_entries = len(specified_dith_only)
    else:
        for identifier in run_names:
            max_entries = max(max_entries, len(list(osbias_err_all[identifier].keys())))

    ncols = 2
    if max_entries == 1:
        ncols = 1
    nrows = int(np.ceil(max_entries / ncols))

    # set up the figure
    plt.clf()
    fig, ax = plt.subplots(nrows, ncols)
    fig.subplots_adjust(hspace=0.4)

    diths_to_consider = []
    if specified_dith_only is not None:
        diths_to_consider = specified_dith_only
    else:
        diths_to_consider = colors.keys()

    for i in range(tot_cases):
        # ----------------------------------------------------------------------------------------
        fsky = fsky_dict[run_names[i]]
        osbias_err = osbias_err_all[run_names[i]]
        row, col = 0, 0
        for dith in diths_to_consider:
            # ----------------------------------------------------------------------------------------
            if dith in osbias_err.keys():
                # ----------------------------------------------------------------------------------------
                # look at the appropriate axis.
                if nrows == 1:
                    if ncols == 1:
                        axis = ax
                    else:
                        axis = ax[col]
                else:
                    axis = ax[row, col]
                # ----------------------------------------------------------------------------------------
                # calcuate the floors with and without shot noise
                if zbin not in w_bao_cls:
                    raise ValueError("Invalid redshift bin: %s" % zbin)
                else:
                    # get the floor with shot noise
                    floor_with_eta = get_stat_floor(
                        ell_arr=ell,
                        zbin=zbin,
                        w_bao_cls_zbin=w_bao_cls[zbin],
                        surf_num_density_zbin=surf_num_density[zbin],
                        dither_strategy=dith,
                        fsky=fsky[dith],
                        with_shot_noise=True,
                    )
                    # get the "best" floor for the fom calculation
                    floor_no_eta = get_stat_floor(
                        ell_arr=ell,
                        zbin=zbin,
                        w_bao_cls_zbin=w_bao_cls[zbin],
                        surf_num_density_zbin=surf_num_density[zbin],
                        dither_strategy=dith,
                        fsky=fsky_best,
                        with_shot_noise=False,
                    )
                # calculate the fo_m
                l = np.arange(np.size(osbias_err[dith]))
                fo_m = get_fom(
                    ell_min,
                    ell_max,
                    l,
                    osbias_err[dith],
                    ell,
                    floor_with_eta,
                    floor_no_eta,
                )
                # ----------------------------------------------------------------------------------------
                if i == 0:
                    # plot the floor with shot noise
                    axis.plot(
                        ell,
                        floor_with_eta,
                        color="k",
                        lw=2.0,
                        label="$\Delta$C$_\ell$: $%s$" % zbin,
                    )

                # set up the legend
                add_leg = ""
                if i == 0:
                    add_leg = "$\mathrm{\sigma_{C_{\ell,OS}}}$: "
                else:
                    add_leg = "         "

                # plot the osbias error
                axis.plot(
                    l,
                    osbias_err[dith],
                    color=colors[run_names[i]],
                    label=r"%s%s ; fo_m: %.6f" % (add_leg, legend_labels[i], fo_m),
                )
                # ----------------------------------------------------------------------------------------

                # plot details
                if (ylim_min is not None) & (ylim_max is not None):
                    axis.set_ylim(ylim_min, ylim_max)
                else:
                    ax.set_ylim(0, 0.00001)
                axis.set_title("%s: $i<%s$" % (dith, lim_mag_i))
                axis.set_xlabel(r"$\ell$")
                axis.set_xlim(0, lmax)
                if tot_cases > 4:
                    leg = axis.legend(
                        labelspacing=0.001,
                    )
                else:
                    leg = axis.legend()
                axis.ticklabel_format(axis="y", style="sci", scilimits=(-2, 2))
                for legobj in leg.legendHandles:
                    legobj.set_linewidth(2.0)
                col += 1
                if col > ncols - 1:
                    col = 0
                    row += 1
    # title to the plot?
    if suptitle is not None:
        plt.suptitle(suptitle, y=1.05)
    # turn off axes on unused panels
    if (max_entries % 2 != 0) and (max_entries > 1):  # i.e. have odd number of diths
        ax[nrows - 1, ncols - 1].axis("off")
    fig.set_size_inches(20, int(nrows * 30 / 7.0))

    # set up the filename
    dith_tag = ""
    if specified_dith_only is not None:
        dith_tag = "%sdith" % max_entries
    date_tag = datetime.date.isoformat(datetime.date.today())
    bias_type_tag = "".join(str(x) for x in filters)
    ell_tag = "_%s<ell<%s" % (ell_min, ell_max)
    if file_append is None:
        file_append = ""

    filename = "%s_OSbiaserr_%s_%s_" % (date_tag, bias_type_tag, dith_tag)
    if tot_cases == 1:
        filename += "%s_" % (run_names[0])
    else:
        filename += "%scadences_" % (tot_cases)
    filename += "th-r<%s_%s_%s_%s_%s%s%s.png" % (
        theory_lim_mag,
        zbin_tag,
        yr_tag,
        poisson_tag,
        zero_pt_tag,
        ell_tag,
        file_append,
    )
    # save the plot
    plt.savefig("%s/%s/%s" % (out_dir, outdir, filename), format="png", bbox_inches="tight")
    print("\nSaved %s" % filename)
    if show_plot:
        plt.show()
    else:
        plt.close("all")



__all__ = ("SNSLMetric",)

import healpy as hp
import numpy as np
from rubin_scheduler.utils import calc_season

import rubin_sim.maf.metrics as metrics
from rubin_sim.maf.utils import collapse_night
from rubin_sim.phot_utils import DustValues


class SNSLMetric(metrics.BaseMetric):
    """Calculate  the number of expected well-measured strongly lensed SN (per data_slice).

    The number of expected strongly lensed SN detections with a well-measured time delay is given by:

    N (lensed SNe Ia with well measured time delay) = 45.7 *
    survey_area / (20000 deg^2) *
    cumulative_season_length / (2.5 years) /
    (2.15 * exp(0.37 * gap_median_all_filter))

    where:
    survey_area: survey area (in deg2)
    cumulative_season_length: cumulative season length (in years)
    gap_median_all_filter: median gap (all filters) (in days)

    (reference? metric originated from Simon Huber and Phillipe Gris)

    Parameters
    ----------
    metricName : str, optional
        metric name
        Default : SNCadenceMetric
    mjd_col : str, optional
        mjd column name
        Default : observationStartMJD,
    filter_col : str, optional
        filter column name
        Default: filter
    night_col : str, optional
        night column name
        Default : night
    m5_col : str, optional
        individual visit five-sigma limiting magnitude (m5) column name
        Default : fiveSigmaDepth
    season: int (list) or -1, optional
        season to process (default: -1: all seasons)
    nfilters_min : int, optional
        The number of filters to demand in a season
        Default: 4.
    min_season_obs : int, optional
        Minimum number of observations per season. Default 5.
    m5mins : dict, optional
        Minimum individual image depth for visit to 'count'.
        Default None uses {'u': 22.7, 'g': 24.1, 'r': 23.7, 'i': 23.1, 'z': 22.2, 'y': 21.4}.
    maps : list, optional
        List of maps to use. Default is the dustmap, to reduce m5 limiting mags accordingly.

    Returns
    -------
    float
        Number of expected well-measured strongly lensed SN
    """

    def __init__(
        self,
        metric_name="SNSLMetric",
        mjd_col="observationStartMJD",
        filter_col="filter",
        night_col="night",
        m5_col="fiveSigmaDepth",
        season=[-1],
        nfilters_min=4,
        min_season_obs=5,
        m5mins=None,
        maps=["DustMap"],
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.filter_col = filter_col
        self.night_col = night_col
        self.m5_col = m5_col
        self.maps = maps

        cols = [self.night_col, self.filter_col, self.mjd_col, self.m5_col]
        super().__init__(col=cols, metric_name=metric_name, maps=self.maps, units="N SL", **kwargs)

        self.bad_val = 0
        self.season = season
        self.bands = "ugrizy"
        if m5mins is None:
            self.m5mins = {
                "u": 22.7,
                "g": 24.1,
                "r": 23.7,
                "i": 23.1,
                "z": 22.2,
                "y": 21.4,
            }
        else:
            self.m5mins = m5mins
        self.min_season_obs = min_season_obs
        self.nfilters_min = nfilters_min
        # Set up dust-extinction values to use to interpret the dust map.
        self.phot_properties = DustValues()

    def n_lensed(self, area, cadence, season_length):
        """
        Parameters
        -----------
        area : float
            Area in square degrees related to this data_slice (sq deg)
        gap_median : float
            median gap between nights with visits (days) - any filter
        cumul_season : float
            length of the season or period of consideration (years)

        Returns
        -------
        float
            Number of strongly lensed SN expected in this area
        """
        # estimate the number of lensed supernovae
        n_lensed_s_ne__ia = 45.7 * area / 20000.0 * season_length / 2.5 / (2.15 * np.exp(0.37 * cadence))
        return n_lensed_s_ne__ia

    def run(self, data_slice, slice_point=None):
        """
        Runs the metric for each data_slice

        Parameters
        ---------------
        data_slice : simulation data
        slice_point:  slice_point(default None)

        Returns
        -----------
        number of SL time delay supernovae

        """
        # If we had no incoming data - just return with badVal.
        if len(data_slice) == 0:
            return self.bad_val

        # Crop it down so things are coadded per night per filter at the median MJD time
        night_slice = collapse_night(
            data_slice,
            night_col=self.night_col,
            filter_col=self.filter_col,
            m5_col=self.m5_col,
            mjd_col=self.mjd_col,
        )
        # Calculate the dust extinction-corrected m5 values and cut visits which don't meet self.m5mins
        for f in np.unique(night_slice[self.filter_col]):
            in_filt = np.where(night_slice[self.filter_col] == f)[0]
            a_x = self.phot_properties.ax1[f] * slice_point["ebv"]
            night_slice[self.m5_col][in_filt] = night_slice[self.m5_col][in_filt] - a_x
            # Set the visits which fall below the minimum to an obvious non-valid value
            night_slice[self.m5_col][in_filt] = np.where(
                night_slice[self.m5_col][in_filt] > self.m5mins[f],
                night_slice[self.m5_col][in_filt],
                -999,
            )
        idxs = np.where(night_slice[self.m5_col] > -998)
        # If nothing survived these cuts, just return with badVal.
        if len(idxs[0]) == 0:
            return self.badval

        # Reset, with coadded per-night/per-filter values, skipping any too-shallow visits.
        night_slice = np.sort(night_slice[idxs], order=self.mjd_col)

        # get the pixel area
        area = hp.nside2pixarea(slice_point["nside"], degrees=True)

        # Note that 'seasons' is the same length as night_slice, and contains integer (season) + float (day)
        seasons = calc_season(np.degrees(slice_point["ra"]), night_slice[self.mjd_col])
        season_ints = np.floor(seasons)

        if self.season == [-1]:
            season_loop = np.unique(season_ints)
        else:
            season_loop = self.season

        n_lensed_s_ne__ia = 0
        for s in season_loop:
            s_idx = np.where(season_ints == s)[0]
            u_filters = np.unique(night_slice[s_idx][self.filter_col])
            if (len(s_idx) < self.min_season_obs) | (np.size(u_filters) < self.nfilters_min):
                # Skip this season
                n_lensed_s_ne__ia += 0
            else:
                # Find the cadence (days) between visits within the season
                cadence = np.diff(night_slice["observationStartMJD"][s_idx])
                # But only the values between nights, not within nights
                cadence = np.median(cadence[np.where(cadence > 0.4)])
                # Season length in years
                season_length = seasons[s_idx][-1] - seasons[s_idx][0]
                n_lensed_s_ne__ia += self.n_lensed(area, cadence, season_length)

        return n_lensed_s_ne__ia



__all__ = (
    "radar",
    "normalize_for_radar",
)

import matplotlib.pylab as plt
import numpy as np
from matplotlib.path import Path
from matplotlib.projections import register_projection
from matplotlib.projections.polar import PolarAxes
from matplotlib.spines import Spine

# Starting with example at
# https://matplotlib.org/examples/api/radar_chart.html


def normalize_for_radar(
    summary,
    norm_run="baseline",
    invert_cols=None,
    reverse_cols=None,
    mag_cols=[],
):
    """
    Normalize values in a dataframe to a given run, return output in a dataframe.

    This provides a similar functionality as the normalize_metric_summaries method, and returns a similar
    dataframe. The options for specifying which columns to invert, reverse, or identify as 'magnitudes'
    are slightly different, instead of using a 'metric_set'.

    Parameters
    ----------
    summary : pandas.DataFrame
        The data frame containing the metric summary stats to normalize (such as from `get_metric_summaries`).
        Note that this should contain only the runs and metrics to be normalized -- e.g.
        `summary.loc[[list of runs], [list of metrics]]`
        summary should be indexed by the run name.
    norm_run : str
        The name of the run to use to define the normalization.
    invert_cols : list of str
        A list of column names that should be inverted (e.g., columns that
        are uncertainties and are better with a smaller value)
    reverse_cols : list of str
        Columns to reverse (e.g., magnitudes)
    mag_cols : list of str
        Columns that are in magnitudes
    """
    out_df = summary.copy()
    if reverse_cols is not None:
        for colname in reverse_cols:
            out_df[colname] = -out_df[colname]
    if invert_cols is not None:
        for colname in invert_cols:
            out_df[colname] = 1.0 / out_df[colname]
    if norm_run is not None:
        indx = np.max(np.where(out_df.index == norm_run)[0])
        for col in out_df.columns:
            if col != "run_name":
                if (col in mag_cols) | (mag_cols == "all"):
                    out_df[col] = 1.0 + (out_df[col] - out_df[col].iloc[indx])
                else:
                    out_df[col] = 1.0 + (out_df[col] - out_df[col].iloc[indx]) / out_df[col].iloc[indx]
    return out_df


def _radar_factory(num_vars, frame="circle"):
    """Create a radar chart with `num_vars` axes.

    This function creates a RadarAxes projection and registers it.

    Parameters
    ----------
    num_vars : int
        Number of variables for radar chart.
    frame : {'circle' | 'polygon'}
        Shape of frame surrounding axes.

    """
    # calculate evenly-spaced axis angles
    theta = np.linspace(0, 2 * np.pi, num_vars, endpoint=False)
    # rotate theta such that the first axis is at the top, then make sure we don't go past 360
    theta += np.pi / 2
    theta = theta % (2.0 * np.pi)

    def draw_poly_patch(self):
        verts = _unit_poly_verts(theta)
        return plt.Polygon(verts, closed=True, edgecolor="k")

    def draw_circle_patch(self):
        # unit circle centered on (0.5, 0.5)
        return plt.Circle((0.5, 0.5), 0.5)

    patch_dict = {"polygon": draw_poly_patch, "circle": draw_circle_patch}
    if frame not in patch_dict:
        raise ValueError("unknown value for `frame`: %s" % frame)

    class RadarAxes(PolarAxes):
        name = "radar"
        # use 1 line segment to connect specified points
        RESOLUTION = 1
        # define draw_frame method
        draw_patch = patch_dict[frame]

        def fill(self, *args, **kwargs):
            """Override fill so that line is closed by default"""
            closed = kwargs.pop("closed", True)
            return super(RadarAxes, self).fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """Override plot so that line is closed by default"""
            lines = super(RadarAxes, self).plot(*args, **kwargs)
            for line in lines:
                self._close_line(line)

        def _close_line(self, line):
            x, y = line.get_data()
            # FIXME: markers at x[0], y[0] get doubled-up
            if x[0] != x[-1]:
                x = np.concatenate((x, [x[0]]))
                y = np.concatenate((y, [y[0]]))
                line.set_data(x, y)

        def set_varlabels(self, labels):
            self.set_thetagrids(np.degrees(theta), labels)

        def _gen_axes_patch(self):
            return self.draw_patch()

        def _gen_axes_spines(self):
            if frame == "circle":
                return PolarAxes._gen_axes_spines(self)
            # The following is a hack to get the spines (i.e. the axes frame)
            # to draw correctly for a polygon frame.

            # spine_type must be 'left', 'right', 'top', 'bottom', or `circle`.
            spine_type = "circle"
            verts = _unit_poly_verts(theta)
            # close off polygon by repeating first vertex
            verts.append(verts[0])
            path = Path(verts)

            spine = Spine(self, spine_type, path)
            spine.set_transform(self.transAxes)
            return {"polar": spine}

    register_projection(RadarAxes)
    return theta


def _unit_poly_verts(theta):
    """Return vertices of polygon for subplot axes.

    This polygon is circumscribed by a unit circle centered at (0.5, 0.5)
    """
    x0, y0, r = [0.5] * 3
    verts = [(r * np.cos(t) + x0, r * np.sin(t) + y0) for t in theta]
    return verts


def radar(
    df,
    rgrids=[0.7, 1.0, 1.3, 1.6],
    colors=None,
    alpha=0.1,
    legend=True,
    figsize=(8.5, 5),
    fill=False,
    bbox_to_anchor=(1.6, 0.5),
    legend_font_size=None,
):
    """
    make a radar plot!
    """
    theta = _radar_factory(np.size(df.columns), frame="polygon")
    fig, axes = plt.subplots(figsize=figsize, subplot_kw=dict(projection="radar"))
    axes.set_rgrids(rgrids, fontsize="x-large")

    if colors is None:
        colors = [None for i in range(len(df))]
    ix = 0
    for i, row in df.iterrows():
        axes.plot(theta, row.values, "o-", label=i, color=colors[ix])
        if fill:
            axes.fill(theta, row.values, alpha=alpha)
        ix += 1

    variables = df.columns.values

    axes.set_varlabels(variables)
    if legend:
        axes.legend(
            bbox_to_anchor=bbox_to_anchor,
            borderaxespad=0,
            loc="lower right",
            fontsize=legend_font_size,
        )
    axes.set_ylim([np.min(rgrids), np.max(rgrids)])

    return fig, axes



from .alm_plots import *
from .artificial_structure_calculation import *
from .coadd_m5_analysis import *
from .constants_for_pipeline import *
from .galaxy_counts_metric_extended import *
from .galaxy_counts_with_pixel_calibration import *
from .masking_algorithm_generalized import *
from .os_bias_analysis import *



import unittest

import numpy as np

from rubin_sim.moving_objects import BaseObs


class TestCamera(unittest.TestCase):
    def setUp(self):
        obj_ra = np.array([10.0, 12.1], float)
        obj_dec = np.array([-30.0, -30.0], float)
        obj_mjd = np.array([59580.16, 59580.16], float)
        self.ephems = np.array(
            list(zip(obj_ra, obj_dec, obj_mjd)),
            dtype=([("ra", float), ("dec", float), ("mjd", float)]),
        )
        obs_ra = np.array([10.0, 10.0], float)
        obs_dec = np.array([-30.0, -30.0], float)
        obs_mjd = np.array([59580.16, 59580.16], float)
        obs_rot_sky_pos = np.zeros(2)
        self.obs = np.array(
            list(zip(obs_ra, obs_dec, obs_rot_sky_pos, obs_mjd)),
            dtype=([("ra", float), ("dec", float), ("rotSkyPos", float), ("mjd", float)]),
        )

    def test_camera_fov(self):
        obs = BaseObs(
            obs_ra="ra",
            obs_dec="dec",
            obs_time_col="mjd",
            footprint="camera",
        )
        idx_obs = obs.sso_in_camera_fov(self.ephems, self.obs)
        self.assertEqual(idx_obs, [0])


if __name__ == "__main__":
    unittest.main()



from .base_obs import *
from .cheby_fits import *
from .cheby_values import *
from .chebyshev_utils import *
from .direct_obs import *
from .ooephemerides import *
from .orbits import *
from .utils import *



"""Run the hourglass metric.
"""
__all__ = ("hourglassPlots",)

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers

from .col_map_dict import col_map_dict


def hourglassPlots(colmap=None, runName="opsim", nyears=10, extraSql=None, extraInfoLabel=None):
    """Run the hourglass metric, for each individual year.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    run_name : `str`, optional
        The name of the simulated survey.
    nyears : `int`, optional
        How many years to attempt to make hourglass plots for.
    extraSql : `str`, optional
        Add an extra sql constraint before running metrics.
    extraInfoLabel : `str`, optional
        Add an extra piece of info_label before running metrics.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    sql = ""
    info_label = ""
    # Add additional sql constraint (such as wfdWhere) and info_label
    if (extraSql is not None) and (len(extraSql) > 0):
        sql = extraSql
        if extraInfoLabel is None:
            info_label = extraSql.replace("filter =", "").replace("filter=", "")
            info_label = info_label.replace('"', "").replace("'", "")
    if extraInfoLabel is not None:
        info_label = extraInfoLabel

    years = list(range(nyears + 1))
    displayDict = {"group": "Hourglass"}
    for year in years[1:]:
        displayDict["subgroup"] = "Year %d" % year
        displayDict["caption"] = (
            "Visualization of the filter usage of the telescope. "
            "The black wavy line indicates lunar phase; the red and blue "
            "solid lines indicate nautical and civil twilight."
        )
        sqlconstraint = "night > %i and night <= %i" % (
            365.25 * (year - 1),
            365.25 * year,
        )
        if len(sql) > 0:
            sqlconstraint = "(%s) and (%s)" % (sqlconstraint, sql)
        md = info_label + " year %i-%i" % (year - 1, year)
        slicer = slicers.HourglassSlicer()
        metric = metrics.HourglassMetric(
            night_col=colmap["night"], mjd_col=colmap["mjd"], metric_name="Hourglass"
        )
        bundle = mb.MetricBundle(
            metric,
            slicer,
            constraint=sqlconstraint,
            info_label=md,
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)



# -*- coding: utf-8 -*-
"""
Created on Tue Sep 25 17:11:20 2018

@author: rstreet
"""

__all__ = ("CalcExpectedVisitsMetric",)

import numpy as np

from rubin_sim.maf.metrics import BaseMetric

from .calculate_lsst_field_visibility_astropy import calculate_lsst_field_visibility


class CalcExpectedVisitsMetric(BaseMetric):
    """Function to calculate the maximum possible number of visits to a
    given pointing, given the expected cadence of observation and within
    the date ranges given, taking target visibility into account.

    Input:
    :param array ra:            RAs, J2000.0, sexigesimal format
    :param array dec:           Decs, J2000.0, sexigesimal format
    :param float cadence:       Interval between successive visits in the
                                same single filter in hours
    :param string start_date:   Start of observing window YYYY-MM-DD
    :param string start_date:   End of observation window YYYY-MM-DD

    Output:
    :param list of arrays n_visits:       Number of visits possible per night
                                          for each pointing
    :param list of arrays hrs_visibility: Hours of visibility per night
                                          for each pointing
    """

    def __init__(
        self,
        pointings,
        cadence,
        start_date,
        end_date,
        filter_id,
        ra_col="fieldRA",
        dec_col="fieldDec",
        metric_name="CalcExpectedVisitsMetric",
        verbose=False,
    ):
        """Input:
        :param array ra:            RAs, J2000.0, sexigesimal format
        :param array dec:           Decs, J2000.0, sexigesimal format
        :param float cadence:       Interval between successive visits in the
                                    same single filter in hours
        :param string start_date:   Start of observing window YYYY-MM-DD
        :param string start_date:   End of observation window YYYY-MM-DD

        Output:
        :param list of arrays n_visits:       Number of visits possible per night
                                              for each pointing
        :param list of arrays hrs_visibility: Hours of visibility per night
                                              for each pointing
        """

        self.pointings = pointings
        self.cadence = cadence
        self.start_date = start_date
        self.end_date = end_date
        self.filter_id = filter_id
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.verbose = verbose

        columns = [self.ra_col, self.dec_col]

        super(CalcExpectedVisitsMetric, self).__init__(col=columns, metric_name=metric_name)

    def run(self, data_slice, slice_point=None):
        n_visits = []
        hrs_visibility = []

        if self.verbose:
            print("Calculating visbility for " + str(len(self.pointings)) + " fields")

        for i in range(0, len(self.pointings), 1):
            # (ra, dec) = pointings[i]
            ra = data_slice[self.ra_col][0]
            dec = data_slice[self.dec_col][0]

            if self.verbose:
                print(" -> RA " + str(ra) + ", Dec " + str(dec))

            (
                total_time_visible,
                hrs_visible_per_night,
            ) = calculate_lsst_field_visibility(ra, dec, self.start_date, self.end_date, verbose=False)

            n_visits.append((np.array(hrs_visible_per_night) / self.cadence).astype(int))
            hrs_visibility.append(np.array(hrs_visible_per_night))

        return n_visits, hrs_visibility



import glob
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import rubin_sim.maf.db as db
import rubin_sim.maf.metric_bundles as metricBundles


def microlensing_fom(
    save_folder,
    result_db_path,
    metric_data_path,
    figsize=None,
    figure_name="microlensing_fom",
):
    """
    Processes a folder, puts together results for
    discovery/detect metric, Npts metric, and Fisher metric,
    and plots them in the four tE bins of 1 - 10 days,
    10 - 30 days, 30 - 100 days, and 100 - 1000 days.

    Parameters
    ----------
    result_db_path : string
        Path to the directory storing the result databases
        generated by MAF.
    metric_data_path : string
        Path to the directory storing the npz files
        generated by MAF.
    """

    # get a dictionary of resultDb from given directory
    result_dbs = get_results_dbs(result_db_path)

    # the following line will be useful if you did not run MAF on all opsims
    run_names = list(result_dbs.keys())

    # retrieve metricBundles for each opsim run and store them in a dictionary
    bundle_dicts = {}

    for run_name in result_dbs:
        bundle_dicts[run_name] = bundle_dict_from_disk(result_dbs[run_name], run_name, metric_data_path)

    # generates results and metric info from default name of file
    results = np.zeros(len(list(bundle_dicts.keys())))
    results_compare = []
    run_names = []
    metric_types = []
    min_t_es = np.zeros(len(list(bundle_dicts.keys())))
    max_t_es = np.zeros(len(list(bundle_dicts.keys())))
    for run in range(len(list(bundle_dicts.keys()))):
        npz = np.load(
            result_db_path + "/" + list(bundle_dicts.keys())[run] + ".npz",
            allow_pickle=True,
        )
        relevant_columns = ["metric_values", "mask"]
        df = pd.DataFrame.from_dict({item: npz[item] for item in relevant_columns})
        run_name, metric_type, min_t_e, max_t_e = parse_t_e_run_types(list(bundle_dicts.keys())[run])
        run_names.append(run_name)
        metric_types.append(metric_type)
        min_t_es[run] = min_t_e
        max_t_es[run] = max_t_e
        results[run] = get_results(df, metric_type)
        if metric_type == "Npts":
            nan_to_be = np.where(df["metric_values"] >= 10e10)[0]
            df["metric_values"][nan_to_be] = np.nan
        results_compare.append(df["metric_values"])
    run_names = np.array(run_names)
    metric_types = np.array(metric_types)
    results_compare = np.array(results_compare)

    plot_fom(
        results,
        run_names,
        metric_types,
        min_t_es,
        max_t_es,
        save_folder,
        figure_name,
        figsize=figsize,
    )
    plot_compare(results_compare, run_names, metric_types, min_t_es, max_t_es, save_folder)

    return


def parse_t_e_run_types(name):
    """
    Parses names of MicrolensingMetric file names

    Parameters
    ----------
    name : string
        A MicrolensingMetric file name
    """
    split_name = name.split("MicrolensingMetric")
    run_name = split_name[0][:-1]
    metric_type = split_name[1].split("_")[1]
    min_t_e = split_name[1].split("_")[3]
    max_t_e = split_name[1].split("_")[4]

    return run_name, metric_type, min_t_e, max_t_e


def get_results(df, run_type, fisher_sigmat_e_t_e_cutoff=0.1):
    """
    Plots the results from the discovery/detect metric, Npts metric,
    and Fisher metric in three sub plots

    Parameters
    ----------
    df : pandas dataframe
        Pandas dataframe of the results npz file
    run_types : numpy array of strings
        Array of strings describing microlensing metric type:
        either 'detect', 'Npts', or 'Fisher' as parsed by the file name
    fisher_sigmat_e_t_e_cutoff : float (Default is 0.1)
        Maximum normalized uncertainty in tE (sigmatE/tE) as determined by
        3sigma values of pubished planet microlensing candidates
    """
    total = len(df)
    if run_type == "detect":
        # Fraction of discovered/detected events
        result = len(np.where(df["metric_values"] == 1)[0]) / total
    elif run_type == "Npts":
        # Average number of points per lightcurve
        result = (
            sum(
                df["metric_values"][~np.isnan(df["metric_values"])][df["metric_values"] >= 0][
                    df["metric_values"] <= 10e10
                ]
            )
            / total
        )
    elif run_type == "Fisher":
        # Fraction of events with sigmatE/tE below the cutoff of 0.1
        result = len(np.where(df["metric_values"] < fisher_sigmat_e_t_e_cutoff)[0]) / total

    return result


def plot_fom(results, run_names, run_types, min_t_e, max_t_e, save_folder, figure_name, figsize):
    """
    Plots the results from the discovery/detect metric, Npts metric,
    and Fisher metric in three sub plots

    Parameters
    ----------
    results : numpy array of floats
        Results from the MicrolensingMetric from get_results() from the
        respective microlensing metric type
    run_names : numpy array of strings
        Array of names of the OpSim run that was used in the metric
    run_types : numpy array of strings
        Array of strings describing microlensing metric type:
        either 'detect', 'Npts', or 'Fisher' as parsed by the file name
    min_t_e : numpy array of int/floats
        Array of values describing the minium einstein crossing time (tE)
        as parsed by the file name
    max_t_e : numpy array of int/floats
        Array of values describing the maximum einstein crossing time (tE)
        as parsed by the file name
    save_folder : string
        String of folder name to save figure
    figure_name : string
        String of figure name
    figsize : tuple
        Tuple of figure size in inches.
        Default is None, which sets figsize = (25, 30)
    """
    if figsize is None:
        figsize = (25, 30)
    fig, (ax1, ax2, ax3) = plt.subplots(1, 3, sharey=True, figsize=figsize)
    plt.tight_layout()
    plt.subplots_adjust(wspace=0, hspace=0)
    subfig_list = [ax1, ax2, ax3]

    plt.rcdefaults()
    font = {"weight": "heavy", "size": 30}

    plt.rc("font", **font)

    t_e_range_list = []
    time_run_names = []
    for i, j in zip(np.unique(min_t_e), np.unique(max_t_e)):
        idx_in_range = np.where((min_t_e >= i) & (max_t_e <= j))
        t_e_range_list.append(idx_in_range)
        time_run_names.append("tE {}-{} days".format(int(i), int(j)))

    detect_runs_idx = np.where(run_types == "detect")
    npts_runs_idx = np.where(run_types == "Npts")
    fisher_runs_idx = np.where(run_types == "Fisher")
    run_type_list = [detect_runs_idx, npts_runs_idx, fisher_runs_idx]

    for t_e_range in range(len(t_e_range_list)):
        for run_type in range(len(run_type_list)):
            # sorted alphabetically according to name of run
            idx_list = list(
                zip(
                    np.intersect1d(t_e_range_list[t_e_range], run_type_list[run_type]),
                    run_names[np.intersect1d(t_e_range_list[t_e_range], run_type_list[run_type])],
                )
            )
            idx_list.sort(key=lambda x: x[1])
            sorted_idxs = np.array([x[0] for x in idx_list])
            subfig_list[run_type].plot(
                results[sorted_idxs],
                run_names[sorted_idxs],
                label=time_run_names[t_e_range],
                marker=".",
                markersize=15,
                linewidth=2.5,
            )

    ax3.legend(bbox_to_anchor=(1, 1), fontsize=20)

    plt.tight_layout()
    plt.subplots_adjust(bottom=0.05)
    ax1.set_xlabel("Discovery Efficiency")
    ax2.set_xlabel("Avg Number of Points")

    ax2.set_xscale("log")
    ax3.set_xlabel("Characaterization Efficiency \n ($\\sigma_{t_E}/t_E$ < 0.1)")
    plt.savefig(save_folder + "/" + figure_name + ".png", bbox_inches="tight")

    return


def plot_compare(results, run_names, run_types, min_t_e, max_t_e, save_folder, npts_required=10):
    """
    Plots confusion matrix type plots comparing fraction detected, characterized (via Fisher),
    and fraction of events with at least npts_required points within 2 tE

    Parameters
    ----------
    results : numpy array of floats
        Results from the MicrolensingMetric from get_results() from the
        respective microlensing metric type
    run_names : numpy array of strings
        Array of names of the OpSim run that was used in the metric
    run_types : numpy array of strings
        Array of strings describing microlensing metric type:
        either 'detect', 'Npts', or 'Fisher' as parsed by the file name
    min_t_e : numpy array of int/floats
        Array of values describing the minium einstein crossing time (tE)
        as parsed by the file name
    max_t_e : numpy array of int/floats
        Array of values describing the maximum einstein crossing time (tE)
        as parsed by the file name
    save_folder : string
        String of folder name to save figures
    npts_required : int (Default is 10).
        Number of poitns within 2tE required for the number of points fraction.
    """

    plt.rcdefaults()
    font = {"weight": "heavy", "size": 20}

    plt.rc("font", **font)

    t_e_range_list = []
    time_run_names = []
    for i, j in zip(np.unique(min_t_e), np.unique(max_t_e)):
        idx_in_range = np.where((min_t_e >= i) & (max_t_e <= j))
        t_e_range_list.append(idx_in_range)
        time_run_names.append("tE {}-{} days".format(int(i), int(j)))

    detect_runs_idx = np.where(run_types == "detect")
    npts_runs_idx = np.where(run_types == "Npts")
    fisher_runs_idx = np.where(run_types == "Fisher")

    run_name_list = np.unique(run_names)

    for t_e_range in range(len(t_e_range_list)):
        for run_name in run_name_list:
            run_name_idxs = np.where(run_names == run_name)
            t_e_run_name_interesct = np.intersect1d(t_e_range_list[t_e_range], run_name_idxs)
            detect_results = results[np.intersect1d(t_e_run_name_interesct, detect_runs_idx)]
            npts_results = results[np.intersect1d(t_e_run_name_interesct, npts_runs_idx)]
            fisher_results = results[np.intersect1d(t_e_run_name_interesct, fisher_runs_idx)]

            detected_fisher_comparison_matrix = detected_fisher_comparison(fisher_results, detect_results)
            fisher_npts_comparison_matrix = fisher_npts_comparison(fisher_results, npts_results)
            detected_npts_comparison_matrix = detected_npts_comparison(detect_results, npts_results)

            confusion_matrix_plot(
                detected_fisher_comparison_matrix,
                "Discovered",
                "Characterized",
                run_name,
                time_run_names[t_e_range],
                save_folder,
            )
            confusion_matrix_plot(
                fisher_npts_comparison_matrix,
                "More than {} Points".format(npts_required),
                "Characterized",
                run_name,
                time_run_names[t_e_range],
                save_folder,
            )
            confusion_matrix_plot(
                detected_npts_comparison_matrix,
                "More than {} Points".format(npts_required),
                "Detected",
                run_name,
                time_run_names[t_e_range],
                save_folder,
            )

    return


def confusion_matrix_plot(comparison_matrix, xlabel, ylabel, run_name, t_e_range, save_folder):
    """
    Plots a confusion matrix type plot comparing two metric types.

    Parameters
    ----------
    comparison_matrix : numpy array
        Array comparing two metric types (A and B) with the following shape:
        [[(Yes A and Yes B), (Yes A and No B)], [(No A and Yes B), (No A and No B)]]
        where Yes A and Yes B are the number of events that pass both the A and B criteria.
    xlabel : string
        Sring of xlabel (also used in file name of figure)
    ylabel : string
        Sring of ylabel (also used in file name of figure)
    run_name : string
        Name of the OpSim run that was used in the metric
        (used in labels and file name)
    t_e_range : string
        String of the range of the tE (used in labels and file name)
    save_folder : string
        String of folder name to save figures
    """

    fig, ax = plt.subplots(figsize=(5, 5))

    ax.matshow(comparison_matrix, cmap=plt.cm.Blues, alpha=0.3)
    for i in range(len(comparison_matrix[0])):
        for j in range(len(comparison_matrix[1])):
            ax.text(
                x=j,
                y=i,
                s="{}".format(comparison_matrix[i, j]),
                va="center",
                ha="center",
                size="medium",
            )

    ax.set(ylabel=ylabel, xlabel=xlabel, title=run_name + "\n" + t_e_range + "\n")
    ax.set_xticklabels([np.nan, "Yes", "No"])
    ax.set_yticklabels([np.nan, "Yes", "No"])
    plt.tight_layout()
    plt.savefig(save_folder + "/{}_{}_{}_{}.png".format(run_name, t_e_range, ylabel, xlabel))
    plt.show()
    plt.close()
    return


def detected_fisher_comparison(fisher_results, detect_results, fisher_sigmat_e_t_e_cutoff=0.1):
    """
    Returns an array of the following form where
    A = fisher criteria and B = detection criteria:
    [[(Yes A and Yes B), (Yes A and No B)], [(No A and Yes B), (No A and No B)]]
    where Yes A and Yes B are the number of events that pass both the A and B criteria.

    Parameters
    ----------
    fisher_results : numpy array
        Array of results from running the Fisher metric of the microlensing metric
    detect_results : numpy array
        Array of results from running the detect metric of the microlensing metric
    fisher_sigmat_e_t_e_cutoff : float (Default is 0.1)
        Maximum normalized uncertainty in tE (sigmatE/tE) as determined by
        3sigma values of pubished planet microlensing candidates
    """
    char_detect = np.where((fisher_results < fisher_sigmat_e_t_e_cutoff) & (detect_results == 1))[0]
    char_ndetect = np.where((fisher_results < fisher_sigmat_e_t_e_cutoff) & (detect_results == 0))[0]
    nchar_detect = np.where((fisher_results > fisher_sigmat_e_t_e_cutoff) & (detect_results == 1))[0]
    nchar_ndetect = np.where((fisher_results > fisher_sigmat_e_t_e_cutoff) & (detect_results == 0))[0]
    return np.array([[len(char_detect), len(char_ndetect)], [len(nchar_detect), len(nchar_ndetect)]])


def fisher_npts_comparison(fisher_results, npts_results, npts_required=10, fisher_sigmat_e_t_e_cutoff=0.1):
    """
    Returns an array of the following form where
    A = fisher criteria and B = npts criteria:
    [[(Yes A and Yes B), (Yes A and No B)], [(No A and Yes B), (No A and No B)]]
    where Yes A and Yes B are the number of events that pass both the A and B criteria.

    Parameters
    ----------
    fisher_results : numpy array
        Array of results from running the Fisher metric of the microlensing metric
    npts_results : numpy array
        Array of results from running the Npts metric of the microlensing metric
    npts_required : int (Default is 10).
        Number of poitns within 2tE required for the number of points fraction.
    fisher_sigmat_e_t_e_cutoff : float (Default is 0.1)
        Maximum normalized uncertainty in tE (sigmatE/tE) as determined by
        3sigma values of pubished planet microlensing candidates
    """
    char_npts = np.where((fisher_results < fisher_sigmat_e_t_e_cutoff) & (npts_results > npts_required))[0]
    char_nnpts = np.where((fisher_results < fisher_sigmat_e_t_e_cutoff) & (npts_results < npts_required))[0]
    nchar_npts = np.where((fisher_results > fisher_sigmat_e_t_e_cutoff) & (npts_results > npts_required))[0]
    nchar_nnpts = np.where((fisher_results > fisher_sigmat_e_t_e_cutoff) & (npts_results < npts_required))[0]
    return np.array([[len(char_npts), len(char_nnpts)], [len(nchar_npts), len(nchar_nnpts)]])


def detected_npts_comparison(detect_results, npts_results, npts_required=10):
    """
    Returns an array of the following form where
    A = detect criteria and B = npts criteria:
    [[(Yes A and Yes B), (Yes A and No B)], [(No A and Yes B), (No A and No B)]]
    where Yes A and Yes B are the number of events that pass both the A and B criteria.

    Parameters
    ----------
    detect_results : numpy array
        Array of results from running the detect metric of the microlensing metric
    npts_results : numpy array
        Array of results from running the Npts metric of the microlensing metric
    npts_required : int (Default is 10).
        Number of poitns within 2tE required for the number of points fraction.
    """
    detect_npts = np.where((detect_results == 1) & (npts_results > npts_required))[0]
    detect_nnpts = np.where((detect_results == 1) & (npts_results < npts_required))[0]
    ndetect_npts = np.where((detect_results == 0) & (npts_results > npts_required))[0]
    ndetect_nnpts = np.where((detect_results == 0) & (npts_results < npts_required))[0]
    return np.array([[len(detect_npts), len(detect_nnpts)], [len(ndetect_npts), len(ndetect_nnpts)]])


def get_results_dbs(result_db_path):
    """
    Create a dictionary of result_db from result_db files
    via PCW Hackathan 2020 Resources

    Args:
        result_db_path(str): Path to the directory storing the result databases
            generated by MAF.

    Returns:
        result_dbs(dict): A dictionary containing the ResultDb objects
            reconstructed from result databases in the provided directory.
    """

    result_dbs = {}
    result_db_list = glob.glob(os.path.join(result_db_path, "*_result.db"))
    for result_db in result_db_list:
        run_name = os.path.basename(result_db).rsplit("_", 1)[0]
        result_db = db.ResultsDb(database=result_db)

        # Don't add empty results.db file,
        if len(result_db.getAllMetricIds()) > 0:
            result_dbs[run_name] = result_db

    return result_dbs


def bundle_dict_from_disk(result_db, run_name, metric_data_path):
    """
    Load metric data from disk and import them into metricBundles.
    via PCW Hackathan 2020 Resources

    Args:
        result_db(dict): A ResultDb object.
        run_name(str): The name of the opsim database that the metrics stored in
            result_db was evaluated on.
        metric_data_path(str): The path to the directory where the metric data
            (.npz files) is stored.

    Returns:
        bundle_dict(dict): A dictionary of metricBundles reconstructed from data
            stored on disk, the keys designate metric names.
    """

    bundle_dict = {}
    display_info = result_db.getMetricDisplayInfo()
    for item in display_info:
        metric_name = item["metric_name"]
        metric_file_name = item["metricDataFile"]
        metric_id = item["metric_id"]
        newbundle = metricBundles.create_empty_metric_bundle()
        newbundle.read(os.path.join(metric_data_path, metric_file_name))
        newbundle.set_run_name(run_name)
        bundle_dict[metric_id, metric_name] = newbundle
    return bundle_dict



__all__ = ("LongGapAGNMetric",)

import numpy as np

from .base_metric import BaseMetric


class LongGapAGNMetric(BaseMetric):
    """max delta-t and average of the top-10 longest gaps."""

    def __init__(
        self,
        metric_name="longGapAGNMetric",
        mjdcol="observationStartMJD",
        units="days",
        xgaps=10,
        badval=-666,
        **kwargs,
    ):
        """Instantiate metric.
        mjdcol = column name for exposure time dates
        """
        cols = [mjdcol]
        super(LongGapAGNMetric, self).__init__(cols, metric_name, units=units, **kwargs)
        self.badval = badval
        self.mjdcol = mjdcol
        self.xgaps = xgaps
        self.units = units

    def run(self, data_slice, slice_point=None):
        metricval = np.diff(data_slice[self.mjdcol])
        return metricval

    def reduce_max_gap(self, metricval):
        if metricval.size > 0:
            result = np.max(metricval)
        else:
            result = self.badval
        return result

    def reduce_average_longest_x_gaps(self, metricval):
        if np.size(metricval) - self.xgaps > 0:
            return np.average(np.sort(metricval)[np.size(metricval) - self.xgaps :])
        else:
            return self.badval



#####################################################################################################
# Purpose: plot skymaps/cartview plots corresponding to alms with specfied l-range (s).
#
# Humna Awan: humna.awan@rutgers.edu
#####################################################################################################
__all__ = ("alm_plots",)

import os

import healpy as hp
import matplotlib.pyplot as plt
import numpy as np


def alm_plots(
    path,
    out_dir,
    bundle,
    nside=128,
    lmax=500,
    filterband="i",
    ra_range=[-50, 50],
    dec_range=[-65, 5],
    subsets_to_consider=[[130, 165], [240, 300]],
    show_plots=True,
):
    """
    Plot the skymaps/cartview plots corresponding to alms with specified l-ranges.
    Automatically creates the output directories and saves the plots.

    Parameters
    ----------
    path : `str`
        path to the main directory where output directory is saved
    out_dir : `str`
        name of the main output directory
    bundle : metricBundle
    nside : `int`
        HEALpix resolution parameter. Default: 128
    lmax : `int`
        upper limit on the multipole. Default: 500
    filterband : `str`
        any one of 'u', 'g', 'r', 'i', 'z', 'y'. Default: 'i'
    ra_range : `np.ndarray`
        range of right ascention (in degrees) to consider in cartview plot; only useful when
                              cartview= True. Default: [-50,50]
    dec_range : `np.ndarray`
        range of declination (in degrees) to consider in cartview plot; only useful when
                               cartview= True. Default: [-65,5]
    subsets_to_consider : `np.ndarray`
        l-ranges to consider, e.g. use [[50, 100]] to consider 50<l<100.
                                                Currently built to handle five subsets (= number of colors built in).
                                                Default: [[130,165], [240, 300]]
    show_plots : `bool`
        set to True if want to show figures. Default: True

    """
    # set up the output directory
    out_dir2 = "almAnalysisPlots_%s<RA<%s_%s<Dec<%s" % (
        ra_range[0],
        ra_range[1],
        dec_range[0],
        dec_range[1],
    )
    if not os.path.exists("%s%s/%s" % (path, out_dir, out_dir2)):
        os.makedirs("%s%s/%s" % (path, out_dir, out_dir2))

    out_dir3 = "almSkymaps"
    if not os.path.exists("%s%s/%s/%s" % (path, out_dir, out_dir2, out_dir3)):
        os.makedirs("%s%s/%s/%s" % (path, out_dir, out_dir2, out_dir3))

    out_dir4 = "almCartviewMaps"
    if not os.path.exists("%s%s/%s/%s" % (path, out_dir, out_dir2, out_dir4)):
        os.makedirs("%s%s/%s/%s" % (path, out_dir, out_dir2, out_dir4))

    # ------------------------------------------------------------------------
    # In order to consider the out-of-survey area as with data=0,  assign the masked region of the
    # skymaps the median of the in-survey data, and then subtract the median off the entire survey.
    # Add the median back later. This gets rid of the massive fake monopole and allows reconstructing
    # the full skymap from components.
    survey_median_dict = {}
    survey_std_dict = {}
    for dither in bundle:
        in_survey = np.where(bundle[dither].metricValues.mask == False)[0]
        out_survey = np.where(bundle[dither].metricValues.mask == True)[0]
        bundle[dither].metricValues.mask[out_survey] = False
        # data pixels
        survey_median = np.median(bundle[dither].metricValues.data[in_survey])
        survey_std = np.std(bundle[dither].metricValues.data[in_survey])
        # assign data[outOfSurvey]= medianData[in_survey]
        bundle[dither].metricValues.data[out_survey] = survey_median
        # subtract median off
        bundle[dither].metricValues.data[:] = bundle[dither].metricValues.data[:] - survey_median
        # save median for later use
        survey_median_dict[dither] = survey_median
        survey_std_dict[dither] = survey_std

    # ------------------------------------------------------------------------
    # now find the alms correponding to the map.
    for dither in bundle:
        array = hp.anafast(
            bundle[dither].metricValues.filled(bundle[dither].slicer.badval),
            alm=True,
            lmax=500,
        )
        cl = array[0]
        alm = array[1]
        l = np.arange(len(cl))

        lsubsets = {}
        color_array = ["y", "r", "g", "m", "c"]
        color = {}
        for case in range(len(subsets_to_consider)):
            lsubsets[case] = (l > subsets_to_consider[case][0]) & (l < subsets_to_consider[case][1])
            color[case] = color_array[case]

        # ------------------------------------------------------------------------
        # plot things out
        plt.clf()
        plt.plot(l, (cl * l * (l + 1)) / (2.0 * np.pi), color="b")
        for key in list(lsubsets.keys()):
            plt.plot(
                l[lsubsets[key]],
                (cl[lsubsets[key]] * l[lsubsets[key]] * (l[lsubsets[key]] + 1)) / (2.0 * np.pi),
                color=color[key],
            )
        plt.title(dither)
        plt.xlabel("$\ell$")
        plt.ylabel(r"$\ell(\ell+1)C_\ell/(2\pi)$")
        filename = "cls_%s.png" % (dither)
        plt.savefig(
            "%s%s/%s/%s" % (path, out_dir, out_dir2, filename),
            format="png",
            bbox_inches="tight",
        )

        if show_plots:
            plt.show()
        else:
            plt.close()

        survey_median = survey_median_dict[dither]
        survey_std = survey_std_dict[dither]

        # ------------------------------------------------------------------------
        # plot full-sky-alm plots first
        n_ticks = 5
        color_min = survey_median - 1.5 * survey_std
        color_max = survey_median + 1.5 * survey_std
        increment = (color_max - color_min) / float(n_ticks)
        ticks = np.arange(color_min + increment, color_max, increment)

        # full skymap
        hp.mollview(
            hp.alm2map(alm, nside=nside, lmax=lmax) + survey_median,
            flip="astro",
            rot=(0, 0, 0),
            min=color_min,
            max=color_max,
            title="",
            cbar=False,
        )
        hp.graticule(dpar=20, dmer=20, verbose=False)
        plt.title("Full Map")

        ax = plt.gca()
        im = ax.get_images()[0]

        fig = plt.gcf()
        cbaxes = fig.add_axes([0.1, 0.015, 0.8, 0.04])  # [left, bottom, width, height]
        cb = plt.colorbar(im, orientation="horizontal", format="%.2f", ticks=ticks, cax=cbaxes)
        cb.set_label("$%s$-band Coadded Depth" % filterband)
        filename = "alm_FullMap_%s.png" % (dither)
        plt.savefig(
            "%s%s/%s/%s/%s" % (path, out_dir, out_dir2, out_dir3, filename),
            format="png",
            bbox_inches="tight",
        )

        # full cartview
        hp.cartview(
            hp.alm2map(alm, nside=nside, lmax=lmax) + survey_median,
            lonra=ra_range,
            latra=dec_range,
            flip="astro",
            min=color_min,
            max=color_max,
            title="",
            cbar=False,
        )
        hp.graticule(dpar=20, dmer=20, verbose=False)
        plt.title("Full Map")
        ax = plt.gca()
        im = ax.get_images()[0]
        fig = plt.gcf()
        cbaxes = fig.add_axes([0.1, -0.05, 0.8, 0.04])  # [left, bottom, width, height]
        cb = plt.colorbar(im, orientation="horizontal", format="%.2f", ticks=ticks, cax=cbaxes)
        cb.set_label("$%s$-band Coadded Depth" % filterband)
        filename = "alm_Cartview_FullMap_%s.png" % (dither)
        plt.savefig(
            "%s%s/%s/%s/%s" % (path, out_dir, out_dir2, out_dir4, filename),
            format="png",
            bbox_inches="tight",
        )

        # prepare for the skymaps for l-range subsets
        color_min = survey_median - 0.1 * survey_std
        color_max = survey_median + 0.1 * survey_std
        increment = (color_max - color_min) / float(n_ticks)
        increment = 1.15 * increment
        ticks = np.arange(color_min + increment, color_max, increment)

        # ------------------------------------------------------------------------
        # consider each l-range
        for case in list(lsubsets.keys()):
            index = []
            low_lim = subsets_to_consider[case][0]
            up_lim = subsets_to_consider[case][1]
            for ll in np.arange(low_lim, up_lim + 1):
                for mm in np.arange(0, ll + 1):
                    index.append(hp.Alm.getidx(lmax=lmax, l=ll, m=mm))
            alms1 = alm.copy()
            alms1.fill(0)
            alms1[index] = alm[index]  # an unmasked array

            # plot the skymap
            hp.mollview(
                hp.alm2map(alms1, nside=nside, lmax=lmax) + survey_median,
                flip="astro",
                rot=(0, 0, 0),
                min=color_min,
                max=color_max,
                title="",
                cbar=False,
            )
            hp.graticule(dpar=20, dmer=20, verbose=False)
            plt.title("%s<$\ell$<%s" % (low_lim, up_lim))
            ax = plt.gca()
            im = ax.get_images()[0]
            fig = plt.gcf()
            cbaxes = fig.add_axes([0.1, 0.015, 0.8, 0.04])  # [left, bottom, width, height]
            cb = plt.colorbar(im, orientation="horizontal", format="%.3f", ticks=ticks, cax=cbaxes)
            cb.set_label("$%s$-band Coadded Depth" % filterband)
            filename = "almSkymap_%s<l<%s_%s.png" % (low_lim, up_lim, dither)
            plt.savefig(
                "%s%s/%s/%s/%s" % (path, out_dir, out_dir2, out_dir3, filename),
                format="png",
                bbox_inches="tight",
            )

            # plot cartview
            hp.cartview(
                hp.alm2map(alms1, nside=nside, lmax=lmax) + survey_median,
                lonra=ra_range,
                latra=dec_range,
                flip="astro",
                min=color_min,
                max=color_max,
                title="",
                cbar=False,
            )
            hp.graticule(dpar=20, dmer=20, verbose=False)
            plt.title("%s<$\ell$<%s" % (low_lim, up_lim))

            ax = plt.gca()
            im = ax.get_images()[0]
            fig = plt.gcf()
            cbaxes = fig.add_axes([0.1, -0.05, 0.8, 0.04])  # [left, bottom, width, height]
            cb = plt.colorbar(im, orientation="horizontal", format="%.3f", ticks=ticks, cax=cbaxes)
            cb.set_label("$%s$-band Coadded Depth" % filterband)
            filename = "almCartview_%s<l<%s_%s.png" % (low_lim, up_lim, dither)
            plt.savefig(
                "%s%s/%s/%s/%s" % (path, out_dir, out_dir2, out_dir4, filename),
                format="png",
                bbox_inches="tight",
            )

        if show_plots:
            plt.show()
        else:
            plt.close("all")



__all__ = ("get_kne_filename", "KnLc", "KNePopMetric", "generate_kn_pop_slicer")

import glob
import os

import numpy as np
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import survey_start_mjd, uniform_sphere

from rubin_sim.maf.metrics import BaseMetric
from rubin_sim.maf.slicers import UserPointsSlicer
from rubin_sim.maf.utils import m52snr
from rubin_sim.phot_utils import DustValues


def get_kne_filename(inj_params_list=None):
    """Given kilonova parameters, get the filename from the grid of models
    developed by M. Bulla

    Parameters
    ----------
    inj_params_list : `list` [`dict`]
        parameters for the kilonova model such as
        mass of the dynamical ejecta (mej_dyn), mass of the disk wind ejecta
        (mej_wind), semi opening angle of the cylindrically-symmetric ejecta
        fan ('phi'), and viewing angle ('theta'). For example
        inj_params_list = [{'mej_dyn': 0.005,
              'mej_wind': 0.050,
              'phi': 30,
              'theta': 25.8}]
    """
    # Get files, model grid developed by M. Bulla
    datadir = get_data_dir()
    file_list = glob.glob(os.path.join(datadir, "maf", "bns", "*.dat"))

    # If no specific parameters passed - return everything.
    if inj_params_list is None or len(inj_params_list) == 0:
        return file_list

    # Otherwise find the parameters for each file and
    # then find the relevant matches.
    params = {}
    matched_files = []
    for filename in file_list:
        key = filename.replace(".dat", "").split("/")[-1]
        params[key] = {}
        params[key]["filename"] = filename
        key_split = key.split("_")
        # Binary neutron star merger models
        if key_split[0] == "nsns":
            mejdyn = float(key_split[2].replace("mejdyn", ""))
            mejwind = float(key_split[3].replace("mejwind", ""))
            phi0 = float(key_split[4].replace("phi", ""))
            theta = float(key_split[5])
            params[key]["mej_dyn"] = mejdyn
            params[key]["mej_wind"] = mejwind
            params[key]["phi"] = phi0
            params[key]["theta"] = theta
        # Neutron star--black hole merger models
        elif key_split[0] == "nsbh":
            mej_dyn = float(key_split[2].replace("mejdyn", ""))
            mej_wind = float(key_split[3].replace("mejwind", ""))
            phi = float(key_split[4].replace("phi", ""))
            theta = float(key_split[5])
            params[key]["mej_dyn"] = mej_dyn
            params[key]["mej_wind"] = mej_wind
            params[key]["phi"] = phi
            params[key]["theta"] = theta
    for key in params.keys():
        for inj_params in inj_params_list:
            match = all([np.isclose(params[key][var], inj_params[var]) for var in inj_params.keys()])
            if match:
                matched_files.append(params[key]["filename"])
                print(f"Found match for {inj_params}")
    print(
        f"Found matches for {len(matched_files)}/{len(inj_params_list)} \
          sets of parameters"
    )

    return matched_files


class KnLc:
    """Read in some KNe lightcurves

    Parameters
    ----------
    file_list : `list` [`str`] or None
        List of file paths to load. If None, loads up all the files
        from data/bns/
    """

    def __init__(self, file_list=None):
        if file_list is None:
            datadir = get_data_dir()
            # Get files, model grid developed by M. Bulla
            file_list = glob.glob(os.path.join(datadir, "maf", "bns", "*.dat"))

        filts = ["u", "g", "r", "i", "z", "y"]
        magidxs = [1, 2, 3, 4, 5, 6]

        # Let's organize the data in to a list of dicts for easy lookup
        self.data = []
        for filename in file_list:
            mag_ds = np.loadtxt(filename)
            t = mag_ds[:, 0]
            new_dict = {}
            for ii, (filt, magidx) in enumerate(zip(filts, magidxs)):
                new_dict[filt] = {"ph": t, "mag": mag_ds[:, magidx]}
            self.data.append(new_dict)

    def interp(self, t, filtername, lc_indx=0):
        """Do the interpolation of the lightcurve for a given time and filter.

        Parameters
        ----------
        t : `np.ndarray`, (N,)
            The times to interpolate the light curve to.
        filtername : `str`
            The filter. one of ugrizy
        lc_index : `int`, optional
            Which file to use.

        Returns
        -------
        result : `np.ndarray`, (N,)
            Array of lightcurve brightnesses at the times of t.=
        """

        result = np.interp(
            t,
            self.data[lc_indx][filtername]["ph"],
            self.data[lc_indx][filtername]["mag"],
            left=99,
            right=99,
        )
        return result


class KNePopMetric(BaseMetric):
    def __init__(
        self,
        metric_name="KNePopMetric",
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        night_col="night",
        pts_needed=2,
        file_list=None,
        mjd0=None,
        output_lc=False,
        badval=-666,
        **kwargs,
    ):
        maps = ["DustMap"]
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.night_col = night_col
        self.pts_needed = pts_needed
        # `bool` variable, if True the light curve will be exported
        self.output_lc = output_lc

        self.lightcurves = KnLc(file_list=file_list)
        self.mjd0 = survey_start_mjd() if mjd0 is None else mjd0

        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1

        cols = [self.mjd_col, self.m5_col, self.filter_col, self.night_col]
        super(KNePopMetric, self).__init__(
            col=cols,
            units="Detected, 0 or 1",
            metric_name=metric_name,
            maps=maps,
            badval=badval,
            **kwargs,
        )

    def _multi_detect(self, around_peak):
        """Simple detection criteria:
        detect at least a certain number of times
        """
        result = 1
        # Detected data points
        if np.size(around_peak) < self.pts_needed:
            return 0

        return result

    def _ztfrest_simple(
        self,
        around_peak,
        mags,
        mags_unc,
        t,
        filters,
        min_dt=0.125,
        min_fade=0.3,
        max_rise=-1.0,
        select_red=False,
        select_blue=False,
    ):
        """Selection criteria based on rise or decay rate;
        simplified version of the methods employed by the ZTFReST project
        (Andreoni & Coughlin et al., 2021)

        Parameters
        ----------
        around_peak : `np.ndarray`, (N,)
            indexes corresponding to 5sigma detections
        mags : `np.ndarray`, (N,)
            magnitudes obtained interpolating models on the data_slice
        t : `np.ndarray`, (N,)
            relative times
        filters : `np.ndarray`, (N,)
            filters in which detections happened
        min_dt : `float`
            minimum time gap between first and last detection in a given band
        min_fade : `float`
            fade rate threshold (positive, mag/day)
        max_rise : `float`
            rise rate threshold (negative, mag/day)
        select_red : `bool`
            if True, only red 'izy' filters will be considered
        select_blue : `bool`
            if True, only blue 'ugr' filters will be considered

        Examples
        ----------
        A transient:
            rising by 0.74 mag/day will pass a threshold max_rise=-0.5
            rising by 0.74 mag/day will not pass a threshold max_rise=-1.0
            fading by 0.6 mag/day will pass a threshold min_fade=0.3
            fading by 0.2 mag/day will not pass a threshold min_fade=0.3
        """
        result = 1

        # Quick check on the number of detected points
        if np.size(around_peak) < self.pts_needed:
            return 0
        # Quick check on the time gap between first and last detection
        elif np.max(t[around_peak]) - np.min(t[around_peak]) < min_dt:
            return 0
        else:
            evol_rate = []
            fil = []
            # Check time gaps and rise or fade rate for each band
            for f in set(filters):
                if select_red is True and f not in "izy":
                    continue
                elif select_blue is True and f not in "ugr":
                    continue
                times_f = t[around_peak][np.where(filters == f)[0]]
                mags_f = mags[around_peak][np.where(filters == f)[0]]
                mags_unc_f = mags_unc[around_peak][np.where(filters == f)[0]]

                # Check if the evolution is significant enough
                idx_max = np.argmax(mags_f)
                idx_min = np.argmin(mags_f)
                if mags_f[idx_min] + mags_unc_f[idx_min] < mags_f[idx_max] - mags_unc_f[idx_max]:
                    signif = True
                else:
                    signif = False

                # Time difference between max and min
                dt_f = np.abs(times_f[idx_max] - times_f[idx_min])

                # Get the evolution rate, if the time gap condition is met
                if dt_f > min_dt and signif is True:
                    # Calculate evolution rate
                    evol_rate_f = (np.max(mags_f) - np.min(mags_f)) / (times_f[idx_max] - times_f[idx_min])
                    evol_rate.append(evol_rate_f)
                else:
                    evol_rate.append(0)
                fil.append(f)
            if len(evol_rate) == 0:
                return 0
            # Check if the conditions on the evolution rate are met
            if np.max(evol_rate) < min_fade and np.min(evol_rate) > max_rise:
                return 0

        return result

    def _multi_color_detect(self, filters):
        """Color-based simple detection criteria:
        detect at least twice, with at least two filters
        """
        result = 1
        # detected in at least two filters
        if np.size(np.unique(filters)) < 2:
            return 0

        return result

    def _red_color_detect(self, filters, min_det=4):
        """Detected at least min_det times in either izy colors

        Parameters
        ----------
        filters : `np.ndarray`, (N,)
            filters in which detections happened
        min_det : `float` or `int`
            minimum number of detections required in izy bands
        """
        result = 1
        # Number of detected points in izy bands
        n_red_det = (
            np.size(np.where(filters == "i")[0])
            + np.size(np.where(filters == "z")[0])
            + np.size(np.where(filters == "y")[0])
        )
        # Condition
        if n_red_det < min_det:
            return 0

        return result

    def _blue_color_detect(self, filters, min_det=4):
        """Detected at least min_det times in either ugr colors

        Parameters
        ----------
        filters : `np.ndarray`, (N,)
            filters in which detections happened
        min_det : `float` or `int`
            minimum number of detections required in ugr bands
        """
        result = 1
        # Number of detected points in ugr bands
        n_blue_det = (
            np.size(np.where(filters == "u")[0])
            + np.size(np.where(filters == "g")[0])
            + np.size(np.where(filters == "r")[0])
        )
        # Condition
        if n_blue_det < min_det:
            return 0

        return result

    def run(self, data_slice, slice_point=None):
        result = {}
        t = data_slice[self.mjd_col] - self.mjd0 - slice_point["peak_time"]
        mags = np.zeros(t.size, dtype=float)

        for filtername in np.unique(data_slice[self.filter_col]):
            infilt = np.where(data_slice[self.filter_col] == filtername)
            mags[infilt] = self.lightcurves.interp(t[infilt], filtername, lc_indx=slice_point["file_indx"])
            # Apply dust extinction on the light curve
            a_x = self.ax1[filtername] * slice_point["ebv"]
            mags[infilt] += a_x

            distmod = 5 * np.log10(slice_point["distance"] * 1e6) - 5.0
            mags[infilt] += distmod

        # Find the detected points
        around_peak = np.where((t > 0) & (t < 30) & (mags < data_slice[self.m5_col]))[0]
        # Filters in which the detections happened
        filters = data_slice[self.filter_col][around_peak]
        # Magnitude uncertainties with Gaussian approximation
        snr = m52snr(mags, data_slice[self.m5_col])
        mags_unc = 2.5 * np.log10(1.0 + 1.0 / snr)

        result["multi_detect"] = self._multi_detect(around_peak)
        result["ztfrest_simple"] = self._ztfrest_simple(
            around_peak, mags, mags_unc, t, filters, select_red=False
        )
        result["ztfrest_simple_red"] = self._ztfrest_simple(
            around_peak, mags, mags_unc, t, filters, select_red=True
        )
        result["ztfrest_simple_blue"] = self._ztfrest_simple(
            around_peak, mags, mags_unc, t, filters, select_blue=True
        )
        result["multi_color_detect"] = self._multi_color_detect(filters)
        result["red_color_detect"] = self._red_color_detect(filters)
        result["blue_color_detect"] = self._blue_color_detect(filters)

        # Export the light curve
        if self.output_lc is True:
            mags[np.where(mags > 50)[0]] = 99.0
            result["lc"] = [
                data_slice[self.mjd_col],
                mags,
                mags_unc,
                data_slice[self.m5_col],
                data_slice[self.filter_col],
            ]
            result["lc_colnames"] = ("t", "mag", "mag_unc", "maglim", "filter")

        return result

    def reduce_multi_detect(self, metric):
        return metric["multi_detect"]

    def reduce_ztfrest_simple(self, metric):
        return metric["ztfrest_simple"]

    def reduce_ztfrest_simple_red(self, metric):
        return metric["ztfrest_simple_red"]

    def reduce_ztfrest_simple_blue(self, metric):
        return metric["ztfrest_simple_blue"]

    def reduce_multi_color_detect(self, metric):
        return metric["multi_color_detect"]

    def reduce_red_color_detect(self, metric):
        return metric["red_color_detect"]

    def reduce_blue_color_detect(self, metric):
        return metric["blue_color_detect"]


def generate_kn_pop_slicer(
    t_start=1,
    t_end=3652,
    n_events=10000,
    seed=42,
    n_files=308,
    d_min=10,
    d_max=300,
    ra=None,
    dec=None,
):
    """Generate a population of KNe events, and put the info about them
    into a UserPointSlicer object

    Parameters
    ----------
    t_start : `float`, optional
        The night to start kilonova events on (days)
    t_end : `float`, optional
        The final night of kilonova events
    n_events : `int`, optional
        The number of kilonova events to generate
    seed : `float`, optional
        The seed passed to np.random
    n_files : `int`, optional
        The number of different kilonova lightcurves to use
        This should match the length of the filenames list passed to the KNePopMetric directly.
    d_min : `float` or `int`, optional
        Minimum luminosity distance (Mpc)
    d_max : `float` or `int`, optional
        Maximum luminosity distance (Mpc)
    ra, dec : `np.ndarray`, (N,) or None
        The ra and dec to use for event positions. Generates uniformly on the spehere if None. (degrees)
    """

    def rndm(a, b, g, size=1):
        """Power-law gen for pdf(x) \propto x^{g-1} for a<=x<=b"""
        r = np.random.random(size=size)
        ag, bg = a**g, b**g
        return (ag + (bg - ag) * r) ** (1.0 / g)

    if ra is None:
        ra, dec = uniform_sphere(n_events, seed=seed)

    peak_times = np.random.uniform(low=t_start, high=t_end, size=n_events)
    file_indx = np.floor(np.random.uniform(low=0, high=n_files, size=n_events)).astype(int)

    # Define the distance
    distance = rndm(d_min, d_max, 4, size=n_events)

    # Set up the slicer to evaluate the catalog we just made
    slicer = UserPointsSlicer(ra, dec, lat_lon_deg=True, badval=0)
    # Add any additional information about each object to the slicer
    slicer.slice_points["peak_time"] = peak_times
    slicer.slice_points["file_indx"] = file_indx
    slicer.slice_points["distance"] = distance

    return slicer



__all__ = (
    "HistogramMetric",
    "AccumulateMetric",
    "AccumulateCountMetric",
    "HistogramM5Metric",
    "AccumulateM5Metric",
    "AccumulateUniformityMetric",
)

import numpy as np
from scipy import stats

from .base_metric import BaseMetric


class VectorMetric(BaseMetric):
    """
    Base for metrics that return a vector
    """

    def __init__(self, bins=None, bin_col="night", col="night", units=None, metric_dtype=float, **kwargs):
        if isinstance(col, str):
            cols = [col, bin_col]
        else:
            cols = list(col) + [bin_col]

        super(VectorMetric, self).__init__(col=cols, units=units, metric_dtype=metric_dtype, **kwargs)
        self.bins = bins
        self.bin_col = bin_col
        self.shape = np.size(bins) - 1


class HistogramMetric(VectorMetric):
    """
    A wrapper to stats.binned_statistic
    """

    def __init__(
        self,
        bins=None,
        bin_col="night",
        col="night",
        units="Count",
        statistic="count",
        metric_dtype=float,
        **kwargs,
    ):
        self.statistic = statistic
        self.col = col
        super(HistogramMetric, self).__init__(
            col=col, bins=bins, bin_col=bin_col, units=units, metric_dtype=metric_dtype, **kwargs
        )

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.bin_col)
        result, bin_edges, bin_number = stats.binned_statistic(
            data_slice[self.bin_col],
            data_slice[self.col],
            bins=self.bins,
            statistic=self.statistic,
        )
        return result


class AccumulateMetric(VectorMetric):
    """
    Calculate the accumulated stat
    """

    def __init__(
        self, col="night", bins=None, bin_col="night", function=np.add, metric_dtype=float, **kwargs
    ):
        self.function = function
        super(AccumulateMetric, self).__init__(
            col=col, bin_col=bin_col, bins=bins, metric_dtype=metric_dtype, **kwargs
        )
        self.col = col

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.bin_col)

        result = self.function.accumulate(data_slice[self.col])
        indices = np.searchsorted(data_slice[self.bin_col], self.bins[1:], side="right")
        indices[np.where(indices >= np.size(result))] = np.size(result) - 1
        result = result[indices]
        result[np.where(indices == 0)] = self.badval
        return result


class AccumulateCountMetric(AccumulateMetric):
    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.bin_col)
        to_count = np.ones(data_slice.size, dtype=int)
        result = self.function.accumulate(to_count)
        indices = np.searchsorted(data_slice[self.bin_col], self.bins[1:], side="right")
        indices[np.where(indices >= np.size(result))] = np.size(result) - 1
        result = result[indices]
        result[np.where(indices == 0)] = self.badval
        return result


class HistogramM5Metric(HistogramMetric):
    """
    Calculate the coadded depth for each bin (e.g., per night).
    """

    def __init__(
        self,
        bins=None,
        bin_col="night",
        m5_col="fiveSigmaDepth",
        units="mag",
        metric_name="HistogramM5Metric",
        **kwargs,
    ):
        super(HistogramM5Metric, self).__init__(
            col=m5_col, bin_col=bin_col, bins=bins, metric_name=metric_name, units=units, **kwargs
        )
        self.m5_col = m5_col

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.bin_col)
        flux = 10.0 ** (0.8 * data_slice[self.m5_col])
        result, bin_edges, bin_number = stats.binned_statistic(
            data_slice[self.bin_col], flux, bins=self.bins, statistic="sum"
        )
        no_flux = np.where(result == 0.0)
        result = 1.25 * np.log10(result)
        result[no_flux] = self.badval
        return result


class AccumulateM5Metric(AccumulateMetric):
    def __init__(
        self, bins=None, bin_col="night", m5_col="fiveSigmaDepth", metric_name="AccumulateM5Metric", **kwargs
    ):
        self.m5_col = m5_col
        super(AccumulateM5Metric, self).__init__(
            bins=bins, bin_col=bin_col, col=m5_col, metric_name=metric_name, **kwargs
        )

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.bin_col)
        flux = 10.0 ** (0.8 * data_slice[self.m5_col])

        result = np.add.accumulate(flux)
        indices = np.searchsorted(data_slice[self.bin_col], self.bins[1:], side="right")
        indices[np.where(indices >= np.size(result))] = np.size(result) - 1
        result = result[indices]
        result = 1.25 * np.log10(result)
        result[np.where(indices == 0)] = self.badval
        return result


class AccumulateUniformityMetric(AccumulateMetric):
    """
    Make a 2D version of UniformityMetric
    """

    def __init__(
        self,
        bins=None,
        bin_col="night",
        exp_mjd_col="observationStartMJD",
        metric_name="AccumulateUniformityMetric",
        survey_length=10.0,
        units="Fraction",
        **kwargs,
    ):
        self.exp_mjd_col = exp_mjd_col
        if bins is None:
            bins = np.arange(0, np.ceil(survey_length * 365.25)) - 0.5
        super(AccumulateUniformityMetric, self).__init__(
            bins=bins, bin_col=bin_col, col=exp_mjd_col, metric_name=metric_name, units=units, **kwargs
        )
        self.survey_length = survey_length

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.bin_col)
        if data_slice.size == 1:
            return np.ones(self.bins.size - 1, dtype=float)

        visits_per_night, blah = np.histogram(data_slice[self.bin_col], bins=self.bins)
        visits_per_night = np.add.accumulate(visits_per_night)
        expected_per_night = np.arange(0.0, self.bins.size - 1) / (self.bins.size - 2) * data_slice.size

        d_max = np.abs(visits_per_night - expected_per_night)
        d_max = np.maximum.accumulate(d_max)
        result = d_max / expected_per_night.max()
        return result



import os
import shutil
import tempfile
import unittest
import warnings

import numpy as np
from rubin_scheduler.data import get_data_dir

from rubin_sim.moving_objects import ChebyFits, Orbits

ROOT = os.path.abspath(os.path.dirname(__file__))


class TestChebyFits(unittest.TestCase):
    def setUp(self):
        self.testdir = os.path.join(get_data_dir(), "tests", "orbits_testdata")
        self.scratch_dir = tempfile.mkdtemp(dir=ROOT, prefix="TestChebyFits-")
        self.orbits = Orbits()
        self.orbits.read_orbits(os.path.join(self.testdir, "test_orbitsMBA.s3m"))
        self.cheb = ChebyFits(
            self.orbits,
            54800,
            30,
            ngran=64,
            sky_tolerance=2.5,
            n_decimal=10,
            n_coeff_position=14,
        )
        self.assertEqual(self.cheb.ngran, 64)

    def tearDown(self):
        del self.orbits
        del self.cheb
        if os.path.exists(self.scratch_dir):
            shutil.rmtree(self.scratch_dir)

    def test_precompute_multipliers(self):
        # Precompute multipliers is done as an automatic step in __init__.
        # After setting up self.cheb, these multipliers should all exist.
        for key in self.cheb.n_coeff:
            self.assertTrue(key in self.cheb.multipliers)

    def test_set_segment_length(self):
        # Expect MBAs with standard ngran and tolerance to have length ~2.0 days.
        self.cheb.calc_segment_length()
        self.assertAlmostEqual(self.cheb.length, 2.0)
        # Test that we can set it to other values which fit into the 30 day window.
        self.cheb.calc_segment_length(length=1.5)
        self.assertEqual(self.cheb.length, 1.5)
        # Test that we if we try to set it to a value which does not fit into the 30 day window,
        # that the actual value used is different - and smaller.
        self.cheb.calc_segment_length(length=1.9)
        self.assertTrue(self.cheb.length < 1.9)
        # Test that we get a warning about the residuals if we try to set the length to be too long.
        with warnings.catch_warnings(record=True) as w:
            self.cheb.calc_segment_length(length=5.0)
            self.assertTrue(len(w), 1)
        # Now check granularity works for other orbit types (which would have other standard lengths).
        # Check for multiple orbit types.
        for orbit_file in [
            "test_orbitsMBA.s3m",
            "test_orbitsOuter.s3m",
            "test_orbitsNEO.s3m",
        ]:
            self.orbits.read_orbits(os.path.join(self.testdir, orbit_file))
            t_start = self.orbits.orbits["epoch"].iloc[0]
            cheb = ChebyFits(self.orbits, t_start, 30, ngran=64, n_decimal=2)
            # And that we should converge for a variety of other tolerances.
            for sky_tolerance in (2.5, 5.0, 10.0, 100.0, 1000.0, 20000.0):
                cheb.sky_tolerance = sky_tolerance
                cheb.calc_segment_length()
                pos_resid, ratio = cheb._test_residuals(cheb.length)
                self.assertTrue(pos_resid < sky_tolerance)
                self.assertEqual((cheb.length * 100) % 1, 0)
                # print('final', orbit_file, sky_tolerance, pos_resid, cheb.length, ratio)
        # And check for challenging 'impactors'.
        for orbit_file in ["test_orbitsImpactors.s3m"]:
            self.orbits.read_orbits(os.path.join(self.testdir, orbit_file))
            t_start = self.orbits.orbits["epoch"].iloc[0]
            cheb = ChebyFits(self.orbits, t_start, 30, ngran=64, n_decimal=10)
            # And that we should converge for a variety of other tolerances.
            for sky_tolerance in (2.5, 10.0, 100.0):
                cheb.sky_tolerance = sky_tolerance
                cheb.calc_segment_length()
                pos_resid, ratio = cheb._test_residuals(cheb.length)
                self.assertTrue(pos_resid < sky_tolerance)
                # print('final', orbit_file, sky_tolerance, pos_resid, cheb.length, ratio)

    @unittest.skip("Skipping because it has a strange platform-dependent failure")
    def test_segments(self):
        # Test that we can create segments.
        self.cheb.calc_segment_length(length=1.0)
        times = self.cheb.make_all_times()
        self.cheb.generate_ephemerides(times)
        self.cheb.calc_segments()
        # We expect calculated coefficients to have the following keys:
        coeff_keys = [
            "obj_id",
            "t_start",
            "t_end",
            "ra",
            "dec",
            "geo_dist",
            "vmag",
            "elongation",
        ]
        for k in coeff_keys:
            self.assertTrue(k in self.cheb.coeffs.keys())
        # And in this case, we had a 30 day timespan with 1 day segments
        # (one day segments should be more than enough to meet 2.5mas tolerance, so not subdivided)
        self.assertEqual(len(self.cheb.coeffs["t_start"]), 30 * len(self.orbits))
        # And we used 14 coefficients for ra and dec.
        self.assertEqual(len(self.cheb.coeffs["ra"][0]), 14)
        self.assertEqual(len(self.cheb.coeffs["dec"][0]), 14)

    def test_write(self):
        # Test that we can write the output to files.
        self.cheb.calc_segment_length()
        self.cheb.generate_ephemerides(self.cheb.make_all_times())
        self.cheb.calc_segments()
        coeff_name = os.path.join(self.scratch_dir, "coeff1.txt")
        resid_name = os.path.join(self.scratch_dir, "resid1.txt")
        failed_name = os.path.join(self.scratch_dir, "failed1.txt")
        self.cheb.write(coeff_name, resid_name, failed_name)
        self.assertTrue(os.path.isfile(coeff_name))
        self.assertTrue(os.path.isfile(resid_name))


class TestRun(unittest.TestCase):
    def setUp(self):
        self.testdir = os.path.join(get_data_dir(), "tests", "orbits_testdata")
        self.scratch_dir = tempfile.mkdtemp(dir=ROOT, prefix="TestChebyFits-")
        self.orbits = Orbits()
        self.orbits.read_orbits(os.path.join(self.testdir, "test_orbitsMBA.s3m"))

    def tearDown(self):
        del self.orbits
        if os.path.exists(self.scratch_dir):
            shutil.rmtree(self.scratch_dir)

    def test_run_through(self):
        # Set up chebyshev fitter.
        t_start = self.orbits.orbits.epoch.iloc[0]
        interval = 30
        cheb = ChebyFits(self.orbits, t_start, interval, ngran=64, sky_tolerance=2.5, n_decimal=10)
        # Set granularity. Use an value that will be too long, to trigger recursion below.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cheb.calc_segment_length(length=10.0)
        # Run through segments.
        cheb.calc_segments()
        self.assertEqual(len(np.unique(cheb.coeffs["obj_id"])), len(self.orbits))
        # Write outputs.
        coeff_name = os.path.join(self.scratch_dir, "coeff2.txt")
        resid_name = os.path.join(self.scratch_dir, "resid2.txt")
        failed_name = os.path.join(self.scratch_dir, "failed2.txt")
        cheb.write(coeff_name, resid_name, failed_name)
        # Test that the segments for each individual object fit together start/end.
        for k in cheb.coeffs:
            cheb.coeffs[k] = np.array(cheb.coeffs[k])
        for obj_id in np.unique(cheb.coeffs["obj_id"]):
            condition = cheb.coeffs["obj_id"] == obj_id
            te_prev = t_start
            for ts, te in zip(cheb.coeffs["t_start"][condition], cheb.coeffs["t_end"][condition]):
                # Test that the start of the current interval = the end of the previous interval.
                self.assertEqual(te_prev, ts)
                te_prev = te
        # Test that the end of the last interval is equal to the end of the total interval
        self.assertEqual(te, t_start + interval)


if __name__ == "__main__":
    unittest.main()



__all__ = ("get_date_version",)

import time
from importlib import metadata


def get_date_version():
    """
    Get today's date and a dictionary with the MAF version information.
    This is written into configuration output files, to help track MAF runs.

    Returns
    -------
    str, dict
        String with today's date, Dictionary with version information.
    """

    version = metadata.version("rubin_sim")
    # today_date = time.strftime("%x")
    today_date = "-".join([time.strftime(x) for x in ["%Y", "%m", "%d"]])
    version_info = {
        "__version__": version,
        "__repo_version__": None,
        "__fingerprint__": None,
        "__dependency_versions__": None,
    }

    return today_date, version_info



__all__ = (
    "NormAirmassStacker",
    "ParallaxFactorStacker",
    "HourAngleStacker",
    "ZenithDistStacker",
    "ParallacticAngleStacker",
    "DcrStacker",
    "FiveSigmaStacker",
    "SaturationStacker",
)

import warnings

import numpy as np
import palpy
from rubin_scheduler.utils import Site, m5_flat_sed

from rubin_sim.maf.utils import load_inst_zeropoints

from .base_stacker import BaseStacker


class SaturationStacker(BaseStacker):
    """Calculate the saturation limit of a point source. Assumes Guassian PSF.

    Parameters
    ----------
    pixscale : float, optional (0.2)
        Arcsec per pixel
    saturation_e : float, optional (150e3)
        The saturation level in electrons
    zeropoints : dict-like, optional (None)
        The zeropoints for the telescope. Keys should be str with filter names, values in mags.
        If None, will use Rubin-like zeropoints.
    km : dict-like, optional (None)
        Atmospheric extinction values.  Keys should be str with filter names.
        If None, will use Rubin-like zeropoints.
    """

    cols_added = ["saturation_mag"]

    def __init__(
        self,
        seeing_col="seeingFwhmEff",
        skybrightness_col="skyBrightness",
        exptime_col="visitExposureTime",
        nexp_col="numExposures",
        filter_col="filter",
        airmass_col="airmass",
        saturation_e=150e3,
        zeropoints=None,
        km=None,
        pixscale=0.2,
    ):
        self.units = ["mag"]
        self.cols_req = [
            seeing_col,
            skybrightness_col,
            exptime_col,
            nexp_col,
            filter_col,
            airmass_col,
        ]
        self.seeing_col = seeing_col
        self.skybrightness_col = skybrightness_col
        self.exptime_col = exptime_col
        self.nexp_col = nexp_col
        self.filter_col = filter_col
        self.airmass_col = airmass_col
        self.saturation_e = saturation_e
        self.pixscale = pixscale
        self.zeropoints = zeropoints
        self.km = km

    def _run(self, sim_data, cols_present=False):
        if self.zeropoints is None:
            zp_inst, k_atm = load_inst_zeropoints()
            self.zeropoints = zp_inst
        if self.km is None:
            self.km = k_atm
        for filtername in np.unique(sim_data[self.filter_col]):
            in_filt = np.where(sim_data[self.filter_col] == filtername)[0]
            # Calculate the length of the on-sky time per EXPOSURE
            exptime = sim_data[self.exptime_col][in_filt] / sim_data[self.nexp_col][in_filt]
            # Calculate sky counts per pixel per second from skybrightness + zeropoint (e/1s)
            sky_counts = (
                10.0 ** (0.4 * (self.zeropoints[filtername] - sim_data[self.skybrightness_col][in_filt]))
                * self.pixscale**2
            )
            # Total sky counts in each exposure
            sky_counts = sky_counts * exptime
            # The counts available to the source (at peak) in each exposure is the
            # difference between saturation and sky
            remaining_counts_peak = self.saturation_e - sky_counts
            # Now to figure out how many counts there would be total, if there are that many in the peak
            sigma = sim_data[self.seeing_col][in_filt] / 2.354
            source_counts = remaining_counts_peak * 2.0 * np.pi * (sigma / self.pixscale) ** 2
            # source counts = counts per exposure (expTimeCol / nexp)
            # Translate to counts per second, to apply zeropoint
            count_rate = source_counts / exptime
            sim_data["saturation_mag"][in_filt] = -2.5 * np.log10(count_rate) + self.zeropoints[filtername]
            # Airmass correction
            sim_data["saturation_mag"][in_filt] -= self.km[filtername] * (
                sim_data[self.airmass_col][in_filt] - 1.0
            )
            # Explicitly make sure if sky has saturated we return NaN
            sim_data["saturation_mag"][np.where(remaining_counts_peak < 0)] = np.nan

        return sim_data


class FiveSigmaStacker(BaseStacker):
    """
    Calculate the 5-sigma limiting depth for a point source in the given conditions.

    This is generally not needed, unless the m5 parameters have been updated
    or m5 was not previously calculated.
    """

    cols_added = ["m5_simsUtils"]

    def __init__(
        self,
        airmass_col="airmass",
        seeing_col="seeingFwhmEff",
        skybrightness_col="skyBrightness",
        filter_col="filter",
        exptime_col="visitExposureTime",
    ):
        self.units = ["mag"]
        self.cols_req = [
            airmass_col,
            seeing_col,
            skybrightness_col,
            filter_col,
            exptime_col,
        ]
        self.airmass_col = airmass_col
        self.seeing_col = seeing_col
        self.skybrightness_col = skybrightness_col
        self.filter_col = filter_col
        self.exptime_col = exptime_col

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is fine.
            return sim_data
        filts = np.unique(sim_data[self.filter_col])
        for filtername in filts:
            infilt = np.where(sim_data[self.filter_col] == filtername)
            sim_data["m5_simsUtils"][infilt] = m5_flat_sed(
                filtername,
                sim_data[infilt][self.skybrightness_col],
                sim_data[infilt][self.seeing_col],
                sim_data[infilt][self.exptime_col],
                sim_data[infilt][self.airmass_col],
            )
        return sim_data


class NormAirmassStacker(BaseStacker):
    """Calculate the normalized airmass for each opsim pointing."""

    cols_added = ["normairmass"]

    def __init__(
        self,
        airmass_col="airmass",
        dec_col="fieldDec",
        degrees=True,
        telescope_lat=-30.2446388,
    ):
        self.units = ["X / Xmin"]
        self.cols_req = [airmass_col, dec_col]
        self.airmass_col = airmass_col
        self.dec_col = dec_col
        self.telescope_lat = telescope_lat
        self.degrees = degrees

    def _run(self, sim_data, cols_present=False):
        """Calculate new column for normalized airmass."""
        # Run method is required to calculate column.
        # Driver runs getColInfo to know what columns are needed from db & which are calculated,
        #  then gets data from db and then calculates additional columns (via run methods here).
        if cols_present:
            # Column already present in data; assume it is correct and does not need recalculating.
            return sim_data
        dec = sim_data[self.dec_col]
        if self.degrees:
            dec = np.radians(dec)
        min_z_possible = np.abs(dec - np.radians(self.telescope_lat))
        min_airmass_possible = 1.0 / np.cos(min_z_possible)
        sim_data["normairmass"] = sim_data[self.airmass_col] / min_airmass_possible
        return sim_data


class ZenithDistStacker(BaseStacker):
    """Calculate the zenith distance for each pointing.
    If 'degrees' is True, then assumes alt_col is in degrees and returns degrees.
    If 'degrees' is False, assumes alt_col is in radians and returns radians.
    """

    cols_added = ["zenithDistance"]

    def __init__(self, alt_col="altitude", degrees=True):
        self.alt_col = alt_col
        self.degrees = degrees
        if self.degrees:
            self.units = ["degrees"]
        else:
            self.unit = ["radians"]
        self.cols_req = [self.alt_col]

    def _run(self, sim_data, cols_present=False):
        """Calculate new column for zenith distance."""
        if cols_present:
            # Column already present in data; assume it is correct and does not need recalculating.
            return sim_data
        if self.degrees:
            sim_data["zenithDistance"] = 90.0 - sim_data[self.alt_col]
        else:
            sim_data["zenithDistance"] = np.pi / 2.0 - sim_data[self.alt_col]
        return sim_data


class ParallaxFactorStacker(BaseStacker):
    """Calculate the parallax factors for each opsim pointing.  Output parallax factor in arcseconds."""

    cols_added = ["ra_pi_amp", "dec_pi_amp"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        date_col="observationStartMJD",
        degrees=True,
    ):
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.date_col = date_col
        self.units = ["arcsec", "arcsec"]
        self.cols_req = [ra_col, dec_col, date_col]
        self.degrees = degrees

    def _gnomonic_project_toxy(self, ra1, dec1, r_acen, deccen):
        """Calculate x/y projection of ra1/dec1 in system with center at r_acen, Deccenp.
        Input radians.
        """
        # also used in Global Telescope Network website
        cosc = np.sin(deccen) * np.sin(dec1) + np.cos(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)
        x = np.cos(dec1) * np.sin(ra1 - r_acen) / cosc
        y = (np.cos(deccen) * np.sin(dec1) - np.sin(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)) / cosc
        return x, y

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct and does not need recalculating.
            return sim_data
        ra_pi_amp = np.zeros(np.size(sim_data), dtype=[("ra_pi_amp", "float")])
        dec_pi_amp = np.zeros(np.size(sim_data), dtype=[("dec_pi_amp", "float")])
        ra_geo1 = np.zeros(np.size(sim_data), dtype="float")
        dec_geo1 = np.zeros(np.size(sim_data), dtype="float")
        ra_geo = np.zeros(np.size(sim_data), dtype="float")
        dec_geo = np.zeros(np.size(sim_data), dtype="float")
        ra = sim_data[self.ra_col]
        dec = sim_data[self.dec_col]
        if self.degrees:
            ra = np.radians(ra)
            dec = np.radians(dec)

        for i, ack in enumerate(sim_data):
            mtoa_params = palpy.mappa(2000.0, sim_data[self.date_col][i])
            # Object with a 1 arcsec parallax
            ra_geo1[i], dec_geo1[i] = palpy.mapqk(ra[i], dec[i], 0.0, 0.0, 1.0, 0.0, mtoa_params)
            # Object with no parallax
            ra_geo[i], dec_geo[i] = palpy.mapqk(ra[i], dec[i], 0.0, 0.0, 0.0, 0.0, mtoa_params)
        x_geo1, y_geo1 = self._gnomonic_project_toxy(ra_geo1, dec_geo1, ra, dec)
        x_geo, y_geo = self._gnomonic_project_toxy(ra_geo, dec_geo, ra, dec)
        # Return ra_pi_amp and dec_pi_amp in arcseconds.
        ra_pi_amp[:] = np.degrees(x_geo1 - x_geo) * 3600.0
        dec_pi_amp[:] = np.degrees(y_geo1 - y_geo) * 3600.0
        sim_data["ra_pi_amp"] = ra_pi_amp
        sim_data["dec_pi_amp"] = dec_pi_amp
        return sim_data


class DcrStacker(BaseStacker):
    """Calculate the RA,Dec offset expected for an object due to differential chromatic refraction.

    For DCR calculation, we also need zenithDistance, HA, and PA -- but these will be explicitly
    handled within this stacker so that setup is consistent and they run in order. If those values
    have already been calculated elsewhere, they will not be overwritten.

    Parameters
    ----------
    filter_col : str
        The name of the column with filter names. Default 'fitler'.
    altCol : str
        Name of the column with altitude info. Default 'altitude'.
    ra_col : str
        Name of the column with RA. Default 'fieldRA'.
    dec_col : str
        Name of the column with Dec. Default 'fieldDec'.
    lstCol : str
        Name of the column with local sidereal time. Default 'observationStartLST'.
    site : str or rubin_sim.utils.Site
        Name of the observory or a rubin_sim.utils.Site object. Default 'LSST'.
    mjdCol : str
        Name of column with modified julian date. Default 'observationStartMJD'
    dcr_magnitudes : dict
        Magitude of the DCR offset for each filter at altitude/zenith distance of 45 degrees.
        Defaults u=0.07, g=0.07, r=0.50, i=0.045, z=0.042, y=0.04 (all arcseconds).

    Returns
    -------
    numpy.array
        Returns array with additional columns 'ra_dcr_amp' and 'dec_dcr_amp' with the DCR offsets
        for each observation.  Also runs ZenithDistStacker and ParallacticAngleStacker.
    """

    cols_added = ["ra_dcr_amp", "dec_dcr_amp"]  # zenithDist, HA, PA

    def __init__(
        self,
        filter_col="filter",
        alt_col="altitude",
        degrees=True,
        ra_col="fieldRA",
        dec_col="fieldDec",
        lst_col="observationStartLST",
        site="LSST",
        mjd_col="observationStartMJD",
        dcr_magnitudes=None,
    ):
        self.units = ["arcsec", "arcsec"]
        if dcr_magnitudes is None:
            # DCR amplitudes are in arcseconds.
            self.dcr_magnitudes = {
                "u": 0.07,
                "g": 0.07,
                "r": 0.050,
                "i": 0.045,
                "z": 0.042,
                "y": 0.04,
            }
        else:
            self.dcr_magnitudes = dcr_magnitudes
        self.zd_col = "zenithDistance"
        self.pa_col = "PA"
        self.filter_col = filter_col
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.degrees = degrees
        self.cols_req = [filter_col, ra_col, dec_col, alt_col, lst_col]
        #  'zenithDist', 'PA', 'HA' are additional columns required, coming from other stackers which must
        #  also be configured -- so we handle this explicitly here.
        self.zstacker = ZenithDistStacker(alt_col=alt_col, degrees=self.degrees)
        self.pastacker = ParallacticAngleStacker(
            ra_col=ra_col,
            dec_col=dec_col,
            mjd_col=mjd_col,
            degrees=self.degrees,
            lst_col=lst_col,
            site=site,
        )
        # Note that RA/Dec could be coming from a dither stacker!
        # But we will assume that coord stackers will be handled separately.

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct and does not need recalculating.
            return sim_data
        # Need to make sure the Zenith stacker gets run first
        # Call _run method because already added these columns due to 'colsAdded' line.
        sim_data = self.zstacker.run(sim_data)
        sim_data = self.pastacker.run(sim_data)
        if self.degrees:
            zenith_tan = np.tan(np.radians(sim_data[self.zd_col]))
            parallactic_angle = np.radians(sim_data[self.pa_col])
        else:
            zenith_tan = np.tan(sim_data[self.zd_col])
            parallactic_angle = sim_data[self.pa_col]
        dcr_in_ra = zenith_tan * np.sin(parallactic_angle)
        dcr_in_dec = zenith_tan * np.cos(parallactic_angle)
        for filtername in np.unique(sim_data[self.filter_col]):
            fmatch = np.where(sim_data[self.filter_col] == filtername)
            dcr_in_ra[fmatch] = self.dcr_magnitudes[filtername] * dcr_in_ra[fmatch]
            dcr_in_dec[fmatch] = self.dcr_magnitudes[filtername] * dcr_in_dec[fmatch]
        sim_data["ra_dcr_amp"] = dcr_in_ra
        sim_data["dec_dcr_amp"] = dcr_in_dec
        return sim_data


class HourAngleStacker(BaseStacker):
    """Add the Hour Angle for each observation.
    Always in HOURS.
    """

    cols_added = ["HA"]

    def __init__(self, lst_col="observationStartLST", ra_col="fieldRA", degrees=True):
        self.units = ["Hours"]
        self.cols_req = [lst_col, ra_col]
        self.lst_col = lst_col
        self.ra_col = ra_col
        self.degrees = degrees

    def _run(self, sim_data, cols_present=False):
        """HA = LST - RA"""
        if cols_present:
            # Column already present in data; assume it is correct and does not need recalculating.
            return sim_data
        if len(sim_data) == 0:
            return sim_data
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            lst = np.radians(sim_data[self.lst_col])
        else:
            ra = sim_data[self.ra_col]
            lst = sim_data[self.lst_col]
        # Check that LST is reasonable
        if (np.min(lst) < 0) | (np.max(lst) > 2.0 * np.pi):
            warnings.warn("LST values are not between 0 and 2 pi")
        # Check that RA is reasonable
        if (np.min(ra) < 0) | (np.max(ra) > 2.0 * np.pi):
            warnings.warn("RA values are not between 0 and 2 pi")
        ha = lst - ra
        # Wrap the results so HA between -pi and pi
        ha = np.where(ha < -np.pi, ha + 2.0 * np.pi, ha)
        ha = np.where(ha > np.pi, ha - 2.0 * np.pi, ha)
        # Convert radians to hours
        sim_data["HA"] = ha * 12 / np.pi
        return sim_data


class ParallacticAngleStacker(BaseStacker):
    """Add the parallactic angle to each visit.
    If 'degrees' is True, this will be in degrees (as are all other angles). If False, then in radians.
    """

    cols_added = ["PA"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        mjd_col="observationStartMJD",
        lst_col="observationStartLST",
        site="LSST",
    ):
        self.lst_col = lst_col
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.degrees = degrees
        self.mjd_col = mjd_col
        self.site = Site(name=site)
        self.units = ["radians"]
        self.cols_req = [self.ra_col, self.dec_col, self.mjd_col, self.lst_col]
        self.ha_stacker = HourAngleStacker(lst_col=lst_col, ra_col=ra_col, degrees=self.degrees)

    def _run(self, sim_data, cols_present=False):
        # Equation from:
        # http://www.gb.nrao.edu/~rcreager/GBTMetrology/140ft/l0058/gbtmemo52/memo52.html
        # or
        # http://www.gb.nrao.edu/GBT/DA/gbtidl/release2pt9/contrib/contrib/parangle.pro
        if cols_present:
            # Column already present in data; assume it is correct and does not need recalculating.
            return sim_data
        # Using the run method (not _run) means that if HA is present, it will not be recalculated.
        sim_data = self.ha_stacker.run(sim_data)
        if self.degrees:
            dec = np.radians(sim_data[self.dec_col])
        else:
            dec = sim_data[self.dec_col]
        sim_data["PA"] = np.arctan2(
            np.sin(sim_data["HA"] * np.pi / 12.0),
            (
                np.cos(dec) * np.tan(self.site.latitude_rad)
                - np.sin(dec) * np.cos(sim_data["HA"] * np.pi / 12.0)
            ),
        )
        if self.degrees:
            sim_data["PA"] = np.degrees(sim_data["PA"])
        return sim_data



__all__ = ("UserPointsSlicer",)

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord

from rubin_sim.maf.plots.spatial_plotters import BaseHistogram, BaseSkyMap

from .base_spatial_slicer import BaseSpatialSlicer


class UserPointsSlicer(BaseSpatialSlicer):
    """A spatial slicer that evaluates pointings overlapping user-provided list of points.

    Parameters
    ----------
    ra : `list` or `numpy.ndarray`
        User-selected RA points, in degrees. Stored internally in radians.
    dec : `list` or `numpy.ndarray`
        User-selected Dec points, in degrees. Stored internally in radians.
    lon_col : `str`, optional
        Name of the longitude (RA equivalent) column to use from the input data.
        Default fieldRA
    lat_col : `str`, optional
        Name of the latitude (Dec equivalent) column to use from the input data.
        Default fieldDec
    latLonDeg : `bool`, optional
        Flag indicating whether the lon and lat values will be in degrees (True) or radians (False).
        Default True (appropriate for opsim v4).
    verbose : `bool`, optional
        Flag to indicate whether or not to write additional information to stdout during runtime.
        Default True.
    badval : `float`, optional
        Bad value flag, relevant for plotting. Default -666.
    leafsize : `int`, optional
        Leafsize value for kdtree. Default 100.
    radius : `float`, optional
        Radius for matching in the kdtree. Equivalent to the radius of the FOV. Degrees.
        Default 2.45.
    use_camera : `bool`, optional
        Flag to indicate whether to use the LSST camera footprint or not.
        Default True.
    camera_footprint_file : `str`, optional
        Name of the camera footprint map to use. Can be None, which will use the default.
    rotSkyPosColName : `str`, optional
        Name of the rotSkyPos column in the input  data. Only used if use_camera is True.
        Describes the orientation of the camera orientation compared to the sky.
        Default rotSkyPos.
    """

    def __init__(
        self,
        ra,
        dec,
        lon_col="fieldRA",
        lat_col="fieldDec",
        lat_lon_deg=True,
        verbose=True,
        badval=-666,
        leafsize=100,
        radius=2.45,
        use_camera=True,
        camera_footprint_file=None,
        rot_sky_pos_col_name="rotSkyPos",
    ):
        super().__init__(
            lon_col=lon_col,
            lat_col=lat_col,
            lat_lon_deg=lat_lon_deg,
            verbose=verbose,
            badval=badval,
            radius=radius,
            leafsize=leafsize,
            use_camera=use_camera,
            camera_footprint_file=camera_footprint_file,
            rot_sky_pos_col_name=rot_sky_pos_col_name,
        )
        # check that ra and dec are iterable, if not, they are probably naked numbers, wrap in list
        if not hasattr(ra, "__iter__"):
            ra = [ra]
        if not hasattr(dec, "__iter__"):
            dec = [dec]
        if len(ra) != len(dec):
            raise ValueError("RA and Dec must be the same length")
        ra = np.radians(ra)
        dec = np.radians(dec)
        self.slice_points["sid"] = np.arange(np.size(ra))
        self.slice_points["ra"] = np.array(ra)
        self.slice_points["dec"] = np.array(dec)
        c = SkyCoord(ra=self.slice_points["ra"] * u.rad, dec=self.slice_points["dec"] * u.rad).transform_to(
            "galactic"
        )
        gall, galb = c.l.rad, c.b.rad
        self.slice_points["gall"] = gall
        self.slice_points["galb"] = galb

        self.nslice = np.size(ra)
        self.shape = self.nslice
        self.slicer_init = {
            "ra": ra,
            "dec": dec,
            "lon_col": lon_col,
            "lat_col": lat_col,
            "radius": radius,
        }
        self.plot_funcs = [BaseSkyMap, BaseHistogram]

    def __eq__(self, other_slicer):
        """Evaluate if two slicers are equivalent."""
        result = False
        # check the slice_points
        for key in other_slicer.slice_points:
            if key in self.slice_points.keys():
                if not np.array_equal(other_slicer.slice_points[key], self.slice_points[key]):
                    return False
            else:
                return False
        if isinstance(other_slicer, UserPointsSlicer):
            if other_slicer.nslice == self.nslice:
                if np.array_equal(
                    other_slicer.slice_points["ra"], self.slice_points["ra"]
                ) and np.array_equal(other_slicer.slice_points["dec"], self.slice_points["dec"]):
                    if other_slicer.lon_col == self.lon_col and other_slicer.lat_col == self.lat_col:
                        if other_slicer.radius == self.radius:
                            if other_slicer.use_camera == self.use_camera:
                                if other_slicer.rotSkyPosColName == self.rotSkyPosColName:
                                    if np.array_equal(other_slicer.shape, self.shape):
                                        result = True
        return result
